# coding: utf-8
#
# BMPC Music Player Client
#
# Copyright (C) 2015 Itaï BEN YAACOV
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import gi

gi.require_version('Gtk', '3.0')

from gi.repository import Gtk, GLib, Gio
import signal
import importlib
import gettext
import os

import ampd

from . import window
from . import shared
from . import __version__
from .utils import omenu
from .utils import actions
from .utils import ssde
from .utils.logger import logger, setup_logfile


class PluginData(object):
    def __init__(self, plugin):
        try:
            self.register_ = plugin.__register__
        except:
            self.register_ = lambda *args: None
        try:
            self.depends = plugin.__depends__
        except:
            self.depends = []
        try:
            self.components = plugin.__components__
        except:
            self.components = []
        try:
            self.menubar_paths = plugin.__menubar_paths__
        except:
            self.menubar_paths = []
        try:
            self.actions = plugin.__actions__
        except:
            self.actions = []

    def register(self, shared):
        self.register_(self, shared)


class App(Gtk.Application, actions.ActionGroup):
    def __init__(self, non_unique=False):
        flags = Gio.ApplicationFlags.NON_UNIQUE if non_unique else 0
        super(App, self).__init__(application_id='begnac.bmpc', flags=flags)

        gettext.install('bmpc', os.path.join(os.path.dirname(__file__), 'locale'))

        self.add_main_option('non-unique', ord('u'), GLib.OptionFlags.NONE, GLib.OptionArg.NONE, _('Do not start a unique instance'), None)
        self.add_main_option(GLib.OPTION_REMAINING, 0, GLib.OptionFlags.NONE, GLib.OptionArg.STRING_ARRAY, '', _('[ACTION...]'))

        self.connect('startup', self.startup_cb)
        self.connect('shutdown', self.shutdown_cb)
        self.connect('handle-local-options', self.handle_local_options_cb)
        self.connect('activate', self.activate_cb)

        self.register()

    @staticmethod
    def startup_cb(self):
        setup_logfile()
        self.shared = shared.AppShared()
        self.shared.menubar_accels_enable = self.menubar_accels_enable
        self.ampd_worker_group = ampd.WorkerGroup(self.shared.ampd_client)
        self.ampd_worker_group.connect_loop(self.idle_output)

        self.notification = Gio.Notification.new(_('MPD status'))
        #self.notification.set_priority(Gio.NotificationPriority.HIGH)
        self.notification_worker = None

        self.add_complex_action('connect', self.shared.ampd_connect)
        self.add_complex_action('disconnect', self.shared.ampd_disconnect)
        self.add_complex_action('profile', self.shared.ampd_connect, parameter_type=GLib.VariantType.new('s'), state=GLib.Variant.new_string(self.shared.config['server']['profile']))
        self.add_complex_action('edit-profiles', self.edit_profiles_cb)

        self.add_complex_action('new-window', self.new_window_cb)
        self.add_complex_action('close-window', self.close_window_cb)
        self.add_complex_action('close-component', self.close_component_cb)
        self.add_complex_action('about', self.about_cb)
        self.add_complex_action('notify', self.worker_hold_app(self.action_notify_cb))
        self.add_complex_action('quit', self.quit_cb)

        self.add_complex_action('play-or-pause', self.worker_hold_app(self.shared.play_or_pause_cb))
        self.add_complex_action('play', self.worker_hold_app(self.shared.mpd_command_cb))
        self.add_complex_action('pause', self.worker_hold_app(self.shared.mpd_command_cb))
        self.add_complex_action('stop', self.worker_hold_app(self.shared.mpd_command_cb))
        self.add_complex_action('next', self.worker_hold_app(self.shared.mpd_command_cb))
        self.add_complex_action('previous', self.worker_hold_app(self.shared.mpd_command_cb))
        self.add_complex_action('fade-to-next', self.worker_hold_app(self.shared.fade_to_next_cb))
        self.add_complex_action('update', self.worker_hold_app(self.shared.mpd_command_cb))

        self.add_complex_action('safe-mode', is_flag=True, shared=self.shared)
        for option_name in self.shared.option_names:
            self.add_complex_action('option-' + option_name, is_flag=True, shared=self.shared)

        self.add_complex_action('BAD', self.THIS_IS_BAD_cb)

        self.shared.connect('notify::safe-mode', self.notify_safe_mode_cb)

        self.output_ids = []
        self.lookup_action('connect').activate()

        self.menubar = omenu.OrderedMenu()
        self.menubar.insert_path('10#bmpc', Gio.MENU_LINK_SUBMENU, _('_BMPC'))
        self.menubar.insert_path('10#bmpc/10#window/10', 'app.new-window', _('New window'), accels=['<Control>n'])
        self.menubar.insert_path('10#bmpc/10#window/20', 'app.close-window', _('Close window'), accels=['<Control>w'])
        self.menubar.insert_path('10#bmpc/10#window/30', 'win.toggle-fullscreen', _('Fullscreen window'), accels=['<Alt>f'])
        self.menubar.insert_path('10#bmpc/20#misc/50', 'app.safe-mode', _('Safe mode'), accels=['<Control><Alt><Shift>s'])
        self.menubar.insert_path('10#bmpc/90#app/50', 'app.quit', _('Quit'), accels=['<Control>q'])
        self.menubar.insert_path('20#edit', Gio.MENU_LINK_SUBMENU, _('_Edit'))
        self.menubar.insert_path('30#playback', Gio.MENU_LINK_SUBMENU, _('_Playback'))
        self.menubar.insert_path('30#playback/10#play/10', 'app.play-or-pause', _('_Play/pause'), accels=['<Control>Up'])
        self.menubar.insert_path('30#playback/10#play/20', 'app.stop', _('_Stop'), accels=['<Control>Down'])
        self.menubar.insert_path('30#playback/20#move/10', 'app.previous', _('_Previous'), accels=['<Control>Left'])
        self.menubar.insert_path('30#playback/20#move/20', 'app.next', _('_Next'), accels=['<Control>Right'])
        self.menubar.insert_path('30#playback/20#move/30', 'app.fade-to-next', _('_Fade to next'), accels=['<Control><Shift>Right'])
        self.menubar.insert_path('40#server', Gio.MENU_LINK_SUBMENU, _('_Server'))
        self.menubar.insert_path('40#server/10#database/10', 'app.update', _('Update database'))
        self.menubar.insert_path('40#server/20#options/10', 'app.option-random', _('Random mode'))
        self.menubar.insert_path('40#server/20#options/20', 'app.option-repeat', _('Repeat mode'))
        self.menubar.insert_path('40#server/20#options/30', 'app.option-consume', _('Consume mode'))
        self.menubar.insert_path('40#server/20#options/40', 'app.option-single', _('Single mode'))
        self.menubar.insert_path('40#server/40#connection/10', 'app.connect', _('Connect'), accels=['<Alt><Shift>c'])
        self.menubar.insert_path('40#server/40#connection/20', 'app.disconnect', _('Disconnect'), accels=['<Alt><Shift>d'])
        self.menubar.insert_path('40#server/40#connection/30#profiles', Gio.MENU_LINK_SUBMENU, _('Profiles'))
        self.menubar.insert_path('40#server/40#connection/30#profiles/20/10', 'app.edit-profiles', _('Edit profiles'))
        self.menubar.insert_path('50#components', Gio.MENU_LINK_SUBMENU, _('_Components'))
        self.menubar.insert_path('50#components/30#current/10', 'app.close-component', _('Close component'), accels=['<Control><Shift>w'])
        self.menubar.insert_path('90#help', Gio.MENU_LINK_SUBMENU, _('_Help'))
        self.menubar.insert_path('90#help/10', 'app.help', _('Help'), hidden_when=None)
        self.menubar.insert_path('90#help/20', 'app.about', _('About'))

        self.app_menu = omenu.OrderedMenu()
        self.app_menu.insert_path('10#window/10', 'app.new-window', _('New window'), accels=['<Control>n'])
        self.app_menu.insert_path('10#window/20', 'app.close-window', _('Close window'), accels=['<Control>w'])
        self.app_menu.insert_path('20#help/10', 'app.help', _('Help'), hidden_when=None)
        self.app_menu.insert_path('20#help/20', 'app.about', _('About'))
        self.app_menu.insert_path('90#quit/10', 'app.quit', _('Quit'), accels=['<Control>q'])
        self.app_menu.insert_path('99#BAD/10', 'app.BAD', 'BAD', accels=['<Control>z'])

        self.component_classes = {}
        self.components = []
        for name in ['current', 'queue', 'browser', 'search', 'savedsearch', 'playlist', 'tanda']:
            self.register_plugin(name)

        self.setup_profile_menu()
        self.set_menubar(self.menubar)
        self.menubar_accels_enable(True)
        self.set_app_menu(self.app_menu)

        self.sigint_source = GLib.unix_signal_add(GLib.PRIORITY_DEFAULT, signal.SIGINT, self.quit_cb)

    @staticmethod
    def shutdown_cb(self):
        logger.debug('Shutting down')
        self.shared.disconnect_by_func(self.notify_safe_mode_cb)
        self.shared.close()
        del self.shared.menubar_accels_enable
        self.clear_actions()
        GLib.source_remove(self.sigint_source)
        for component in self.components:
            component.close()

    @staticmethod
    def handle_local_options_cb(self, options):
        options = options.end().unpack()
        if GLib.OPTION_REMAINING in options:
            for action in options[GLib.OPTION_REMAINING]:
                self.activate_action(action)
            return 0
        return -1

    @staticmethod
    def activate_cb(self, *args):
        win = self.get_active_window()
        if win:
            win.present()
        else:
            self.new_window_cb()

    def worker_hold_app(self, f):
        def g(*args, **kwargs):
            retval = f(*args, **kwargs)
            if isinstance(retval, ampd.Worker):
                self.hold()
                retval.add_atexit(self.release)
            return retval
        return g

    def menubar_accels_enable(self, enable):
        for action, accels in self.menubar.get_accels().items():
            self.set_accels_for_action(action, accels if enable else [])

    def register_plugin(self, name):
        if name in self.shared.plugins:
            return

        plugin_data = self.shared.plugins[name] = PluginData(importlib.import_module('bmpc.plugins.' + name))
        for dep_name in plugin_data.depends:
            self.register_plugin(dep_name)

        plugin_data.register(self.shared)
        for component in plugin_data.components:
            self.register_component_class(component)
        for path in plugin_data.menubar_paths:
            self.menubar.insert_path(*path)
        for name, cb in plugin_data.actions:
            self.add_complex_action(name, self.worker_hold_app(cb))

    def unregister_plugin(self, name):
        raise NotImplementedError

    def setup_profile_menu(self):
        menu = self.menubar.insert_path('40#server/40#connection/30#profiles/10#profiles', Gio.MENU_LINK_SECTION)
        menu.remove_all()
        for profile in self.shared.config['server']['profiles']:
            menu.append(profile['name'], 'app.profile("{}")'.format(profile['name']))

    def pop_named_component(self, name):
        for component in self.components:
            if component.name == name:
                self.components.remove(component)
                return component
        return None

    def start_component(self, name, new):
        component = (not new and self.pop_named_component(name)) or self.component_classes[name](self.shared)
        self.components.append(component)
        if component.win:
            win = component.win
            component.restart()
        else:
            win = (not new and self.get_active_window()) or window.Window(self)
            win.change_component(component)
        win.present()

        if False:
            for component in self.components:
                if not component.win:
                    self.components.remove(component)
                    component.close()

    def action_component_activate_cb(self, action, parameter):
        name = action.get_name()
        if name.endswith('-newinstance'):
            self.start_component(name[:-len('-newinstance')], True)
        else:
            self.start_component(name, False)

    def register_component_class(self, component_class):
        if self.get_is_registered() and not self.get_is_remote():
            self.component_classes[component_class.name] = component_class
            self.add_complex_action(component_class.name, self.action_component_activate_cb)
            self.add_complex_action(component_class.name + '-newinstance', self.action_component_activate_cb)
            self.menubar.insert_path('50#components/10#components/' + component_class.key,
                                     'app.' + component_class.name,
                                     component_class.title,
                                     accels=['<Alt>' + component_class.key])
            self.menubar.insert_path('50#components/20#components-newinstance/' + component_class.key,
                                     'app.' + component_class.name + '-newinstance',
                                     component_class.title + _(' (new instance)'),
                                     accels=['<Ctrl><Alt>' + component_class.key])

    def new_window_cb(self, *args):
        self.start_component('current', True)

    def close_window_cb(self, *args):
        self.get_active_window().destroy()

    def close_component_cb(self, *args):
        win = self.get_active_window()
        old_component = win.component
        for component in self.components:
            if not component.win:
                win.change_component(component)
                break
        else:
            win.destroy()
        old_component.close()
        self.components.remove(old_component)

    def notify_safe_mode_cb(self, *args):
        if self.shared.safe_mode:
            for option in self.shared.option_names:
                self.shared.set_property('option-' + option, False)

    def quit_cb(self, *args):
        logger.debug('Quit')
        for win in self.get_windows():
            win.destroy()
        return GLib.SOURCE_CONTINUE

    def about_cb(self, *args):
        dialog = Gtk.AboutDialog(parent=self.get_active_window(), program_name='BMPC Music Player Client', version=__version__, comments=_('A Python/Gtk3 client for MPD'), copyright='© 2015 Itaï BEN YAACOV', license_type=Gtk.License.GPL_3_0, logo_icon_name='face-cool', website='http://math.univ-lyon1.fr/~begnac', website_label=_('Author\'s website'))
        dialog.run()
        dialog.destroy()

    def edit_profiles_cb(self, *args):
        struct = ssde.List(
            label=_('Profiles'),
            substruct=ssde.Dict(
                label=_('Profile'),
                substructs=[
                    ssde.Text(name='name', label=_('Profile name')),
                    ssde.Text(name='host', label=_('Host name')),
                    ssde.Integer(name='port', label=_('Port number'), default=6600),
                ]))
        value = struct.dialog(self.get_active_window(), self.shared.config['server']['profiles'], self.shared.config.setdefault('server', {}).setdefault('edit_dialog_size', [500,500]))
        if value:
            self.shared.config['server']['profiles'] = value
            self.setup_profile_menu()

    @ampd.worker
    def action_notify_cb(self, *args):
        if self.notification_worker:
            self.notification_worker.close()
            self.withdraw_notification('status')
        self.notification_worker = yield ampd._self()
        yield ampd.IDLE
        if self.shared.state == 'stop':
            icon_name = 'media-playback-stop-symbolic'
            body = 'Stopped'
        else:
            if self.shared.state == 'play':
                icon_name = 'media-playback-start-symbolic'
            else:
                icon_name = 'media-playback-pause-symbolic'
            body = '{0} / {1}'.format(self.shared.current_song.get('artist', '???'), self.shared.current_song.get('title', '???'))
            if 'performer' in self.shared.current_song:
                body += ' / ' + self.shared.current_song['performer']
        self.notification.set_body(body)
        self.notification.set_icon(Gio.Icon.new_for_string(icon_name))
        self.send_notification('status', self.notification)
        yield ampd.TIMEOUT(5000)
        self.withdraw_notification('status')
        self.notification_worker = None

    @ampd.worker
    def idle_output(self):
        for output_id in self.output_ids:
            self.remove_action('output.' + output_id)
        self.output_ids = []
        outputs = yield ampd.outputs()
        for output in outputs:
            output_id = output['outputid']
            self.output_ids.append(output_id)
            self.menubar.insert_path('40#server/30#output', Gio.MENU_LINK_SECTION).remove_all()
            self.menubar.insert_path('40#server/30#output/' + output_id, 'app.output.' + output_id, output['outputname'])
            self.add_complex_action('output.' + output_id, self.action_output_activate_cb, state=GLib.Variant.new_boolean(int(output['outputenabled'])))
        while True:
            yield ampd.OUTPUT
            outputs = yield ampd.outputs()
            for output in outputs:
                self.lookup_action('output.' + output['outputid']).set_state(GLib.Variant.new_boolean(int(output['outputenabled'])))

    @ampd.worker
    def action_output_activate_cb(self, action, parameter):
        yield ampd.toggleoutput(action.get_name().split('.', 1)[1])

    def THIS_IS_BAD_cb(self, *args):
        self.find_accels()
        #self.worker_hold_app(self.BAD2)()
        #self.get_active_window().scale_time_break_interaction()

    @ampd.worker
    def BAD2(self):
        yield ampd.CONNECT
        yield ampd.addid('rrr')

    def find_accels(self):
        actions = self.list_actions()
        print(actions)
        accels = {}
        for action in actions:
            accels[action] = self.get_accels_for_action(action)
        print(accels)
