# coding: utf-8
#
# BMPC Music Player Client
#
# Copyright (C) 2015 Itaï BEN YAACOV
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from gi.repository import GObject, Gtk, Pango
import re


def store_set_rows(store, rows, func):
    if not rows:
        store.clear()
        return
    i = store.get_iter_first()
    appending = (i == None)
    for row in rows:
        if appending:
            i = store.append()
        func(i, row)
        if not appending:
            i = store.iter_next(i)
            if i == None:
                appending = True
    if not appending:
        while store.remove(i):
            pass


def config_notify_cb(obj, param, config):
    config[param.name] = obj.get_property(param.name)


class Field(GObject.GObject):
    title = GObject.property(type=str)
    width = GObject.property(type=int)
    visible = GObject.property(type=bool, default=True)

    get_value = None

    def __init__(self, name, title=None, min_width=50, visible=True, get_value=None):
        super(Field, self).__init__()
        self.name = name
        self.title = title
        self.width = self.min_width = min_width
        self.visible = visible
        if get_value:
            self.get_value = get_value

    def get_renderer(self):
        return Gtk.CellRendererText(ellipsize=Pango.EllipsizeMode.END, background_set=True)


class FieldWithTable(Field):
    def __init__(self, name, title=None, table=None, width=50, visible=True):
        super(FieldWithTable, self).__init__(name, title, width, visible)
        self.table = table

    def get_value(self, song):
        for field, pattern, value in self.table:
            if not field:
                return value
            else:
                match = re.search(pattern, song.get(field, ''))
                if match:
                    return match.expand(value)


class FieldFamily(GObject.GObject):
    order = GObject.property()

    def __init__(self, config):
        super(FieldFamily, self).__init__()
        self.config = config
        self.old_info = self.config.get('info', {})
        self.old_order = self.config.get('order', [])
        self.config['info'] = {}
        self.order = self.config['order'] = []
        self.connect('notify', config_notify_cb, self.config)

        self.names = []
        self.basic_names = []
        self.derived_names = []
        self.fields = {}

    def close(self):
        self.disconnect_by_func(config_notify_cb)

    def register_field(self, field):
        if field.name in self.names:
            raise RuntimeError('Field "{}" already registered'.format(field.name))
        self.names.append(field.name)
        if field.get_value:
            self.derived_names.append(field.name)
        else:
            self.basic_names.append(field.name)
        self.fields[field.name] = field

        field_config = self.config['info'].setdefault(field.name, self.old_info.get(field.name, {}))
        field.connect('notify', config_notify_cb, field_config)
        field.width = field_config.setdefault('width', field.width)
        field.visible = field_config.setdefault('visible', field.visible)
        if field.name in self.order:
            return
        if not field.name in self.old_order:
            self.order.append(field.name)
            return
        pos = self.old_order.index(field.name)
        for i, name in enumerate(self.order):
            if name not in self.old_order or self.old_order.index(name) > pos:
                self.order.insert(i, field.name)
                return
        else:
            self.order.append(field.name)

    def unregister_field(self, field):
        self.names.remove(field.name)
        if field.name in self.basic_names:
            self.basic_names.remove(field.name)
        else:
            self.derived_names.remove(field.name)
        del self.fields[field.name]
        field.disconnect_by_func(config_notify_cb)

    def song_set_fields(self, song):
        for name in self.derived_names:
            value = self.fields[name].get_value(song)
            if value is not None:
                song[name] = value

    def songs_set_fields(self, songs):
        for song in songs:
            self.song_set_fields(song)


class RowDict(GObject.Object):
    def __init__(self, data):
        super(RowDict, self).__init__()
        self.set_data(data)

    def set_data(self, data):
        self.__dict__['_data'] = data

    def get_data(self):
        return self._data

    def __getattr__(self, name):
        return self._data.get(name)

    def __setattr__(self, name, value):
        self._data[name] = value

    def __delattr__(self, name):
        self._data.pop(name, None)


class StoreWithIters(object):
    def iters(self):
        i = self.get_iter_first()
        while i:
            yield i
            i = self.iter_next(i)


class DictStore(StoreWithIters, Gtk.ListStore):
    def __init__(self):
        super(DictStore, self).__init__(RowDict)

    def get_row(self, i):
        return self.get_value(i, 0)

    def set_row(self, i, row):
        self.set_value(i, 0, RowDict(row))

    def set_rows(self, rows):
        store_set_rows(self, rows, self.set_row)


class DictStoreSort(StoreWithIters, Gtk.TreeModelSort):
    def __init__(self):
        self.store = DictStore()
        super(DictStoreSort, self).__init__(self.store)

    def get_row(self, i):
        return self.store.get_row(self.convert_iter_to_child_iter(i))

    def set_rows(self, rows):
        self.store.set_rows(rows)


class DictTreeView(Gtk.TreeView):
    def __init__(self, fields, data_func, sortable):
        super(DictTreeView, self).__init__(model=DictStoreSort() if sortable else DictStore(),visible=True, vexpand=True, enable_grid_lines=Gtk.TreeViewGridLines.BOTH, rubber_banding=True, search_column=0)
        self.fields = fields
        self.sortable = sortable
        self.cols = {name: FieldColumn(self.fields.fields[name], data_func) for name in self.fields.order}
        self.changing_columns = False

        self.set_search_equal_func(lambda store, col, key, i: not any(isinstance(value, str) and key.lower() in value.lower() for value in store.get_row(i).get_data().values()))

        if self.sortable:
            store = self.get_model()
            for i, name in enumerate(self.fields.order):
                self.cols[name].set_sort_column_id(i)
                store.set_sort_func(i, self.sort_func, name)

        self.fields_notify_order_cb()
        self.connect('destroy', self.destroy_cb)
        self.connect('columns-changed', self.columns_changed_cb)
        self.fields.connect('notify::order', self.fields_notify_order_cb)

    @staticmethod
    def destroy_cb(self):
        self.fields.disconnect_by_func(self.fields_notify_order_cb)
        self.disconnect_by_func(self.columns_changed_cb)
        del self.cols

    @staticmethod
    def sort_func(store, i, j, name):
        try:
            v1 = store.get_row(i).get_data().get(name)
            v2 = store.get_row(j).get_data().get(name)
            return 0 if v1 == v2 else -1 if v1 is None or (v2 is not None and v1 < v2) else 1
        except:
            return 0

    @staticmethod
    def columns_changed_cb(self):
        if not self.changing_columns:
            self.fields.order = [self.get_column(i).field.name for i in range(self.get_n_columns())]

    def fields_notify_order_cb(self, *args):
        self.changing_columns = True  #### BLOCK ??
        for i, name in enumerate(self.fields.order):
            col = self.cols[name]
            if col is not self.get_column(i):
                if col.get_tree_view():
                    self.remove_column(col)
                self.insert_column(col, i)
        self.changing_columns = False

    def get_selection_rows(self):
        store, paths = self.get_selection().get_selected_rows()
        return [store.get_row(store.get_iter(p)).get_data() for p in paths], [Gtk.TreeRowReference.new(store, p) for p in paths]


class FieldColumn(Gtk.TreeViewColumn):
    width_rw = GObject.property(type=int)

    def __init__(self, field, next_data_func):
        super(FieldColumn, self).__init__()
        self.field = field
        self.renderer = field.get_renderer()
        self.next_data_func = next_data_func
        field.bind_property('title', self, 'title', GObject.BindingFlags.SYNC_CREATE)
        field.bind_property('visible', self, 'visible', GObject.BindingFlags.SYNC_CREATE)
        self.set_min_width(field.min_width)
        self.connect('notify::width-rw', self.notify_width_rw_cb)
        field.bind_property('width', self, 'width-rw', GObject.BindingFlags.SYNC_CREATE)
        self.bind_property('width', field, 'width')
        self.set_resizable(True)
        self.set_reorderable(True)
        self.pack_start(self.renderer, True)
        self.set_cell_data_func(self.renderer, self.data_func)

    @staticmethod
    def notify_width_rw_cb(self, detail):
        if self.width_rw != self.get_width():
            self.set_fixed_width(self.width_rw)

    @staticmethod
    def data_func(self, renderer, store, i, arg):
        value = store.get_row(i).get_data().get(self.field.name)
        renderer.set_property('text', str(value) if value is not None else None)
        return self.next_data_func and self.next_data_func(self, renderer, store, i, arg)
