# coding: utf-8
#
# BMPC Music Player Client
#
# Copyright (C) 2015 Itaï BEN YAACOV
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from gi.repository import Gtk, GLib

import ampd

from . import songlist


class PlayQueue(songlist.SongList):
    title = _('Play Queue')
    name = 'queue'
    key = '1'

    duplicate_test_columns = ['title']

    context_menu_paths = [('90/10', 'comp.shuffle', _('Shuffle'))]

    def __init__(self, *args, **kwargs):
        super(PlayQueue, self).__init__(*args, **kwargs)
        self.actions.add_complex_action('shuffle', self.action_shuffle_cb, unsafe=True, shared=self.shared)
        self.notify_current_song_handler = self.shared.connect('notify::current-song', self.notify_current_song_cb)
        self.current_song_id = None
        for action in self.actions.list_actions():
            if action.startswith('queue-'):
                self.actions.remove(action)

    def close(self):
        self.shared.disconnect(self.notify_current_song_handler)
        super(PlayQueue, self).close()

    def connect_cb(self):
        self.idle_playlist()

    @ampd.worker
    def idle_playlist(self):
        while True:
            self.set_songs((yield ampd.playlistinfo()))
            if not self.current_song_id:
                self.notify_current_song_cb()
            yield ampd.PLAYLIST

    def data_func(self, column, renderer, store, i, j):
        super(PlayQueue, self).data_func(column, renderer, store, i, j)
        if self.shared.state != 'stop' and store.get_row(i).id == self.current_song_id:
            renderer.set_property('font', 'bold')
        else:
            renderer.set_property('font', None)

    @ampd.worker
    def action_shuffle_cb(self, *args):
        yield ampd.shuffle()

    def notify_current_song_cb(self, *args):
        new_id = self.shared.current_song.get('id')
        if new_id and new_id != self.current_song_id:
            p = Gtk.TreePath.new_from_string(self.shared.current_song['pos'])
            self.treeview.set_cursor(p)
            self.treeview.scroll_to_cell(p, None, True, 0.5, 0.0)
        self.current_song_id = new_id
        self.treeview.queue_draw()

    def new_song(self, i):
        index = self.store.get_path(i).get_indices()[0]
        ampd.worker(lambda x: (yield ampd.addid(self.store.get_row(i).file, index)))(self)
        return True

    def delete_song(self, i):
        song_id = self.store.get_row(i).id
        m = int(self.store.get_string_from_iter(i))
        for n in range(self.store.iter_n_children()):
            if n != m and self.store.get_row(self.store.iter_nth_child(None, n)).id == song_id:
                ampd.worker(lambda x: (yield ampd.swap(n, m), ampd.delete(m)))(self)
                return True
        if self.shared.safe_mode and self.shared.state == 'play' and self.shared.current_song.get('id') == song_id:
            return False
        else:
            ampd.worker(lambda x: (yield ampd.deleteid(song_id)))(self)
            return True

    @ampd.worker
    def treeview_row_activated_cb(self, treeview, p, column):
        if self.shared.safe_mode:
            return
        yield ampd.playid(self.store.get_row(self.store.get_iter(p)).id)


@ampd.worker
def action_queue_add_replace_cb(songlist, action, parameter):
    filenames = songlist.get_filenames(not parameter.get_boolean())
    replace = '-replace' in action.get_name()
    shuffle = '-shuffle' in action.get_name()
    if replace:
        yield ampd.clear()
    yield (ampd.add(filename) for filename in filenames)
    if shuffle:
        yield ampd.shuffle()
    if replace:
        yield ampd.play()


__components__ = [PlayQueue]
__depends__ = ['songlist']

def __register__(plugin_data, shared):
    songlist_shared = shared.plugins['songlist']
    songlist_shared.actions += [('queue' + verb + shuffle, action_queue_add_replace_cb,
                                 {'unsafe': verb == '-replace' or shuffle, 'parameter_type': GLib.VariantType.new('b')})
                                for verb in ('-add', '-replace')
                                for shuffle in ('', '-shuffle')]
    for paths, parameter in ((songlist_shared.context_menu_paths, '(false)'), (songlist_shared.left_context_menu_paths, '(true)')):
        paths.append(('10/1', 'comp.queue-add' + parameter, _('Add to queue')))
        paths.append(('10/2', 'comp.queue-replace' + parameter, _('Replace queue')))
        paths.append(('15/1', 'comp.queue-add-shuffle' + parameter, _('Add to queue & shuffle')))
        paths.append(('15/2', 'comp.queue-replace-shuffle' + parameter, _('Replace queue & shuffle')))
