# coding: utf-8
#
# BMPC Music Player Client
#
# Copyright (C) 2015 Itaï BEN YAACOV
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from gi.repository import Gtk

import ampd

from bmpc.utils import ssde

from . import songlist


class SavedSearch(songlist.PanedComponent, songlist.SongListWithSeparator):
    title = _('Saved Searches')
    name = 'savedsearch'
    key = '4'

    duplicate_test_columns = ['title', 'date']
    sortable = True

    left_context_menu_paths = [('50/10', 'comp.search-new', _('New search')),
                               ('50/20', 'comp.search-edit', _('Edit search')),
                               ('50/30', 'comp.search-delete', _('Delete search'))]

    def __init__(self, *args, **kwargs):
        super(SavedSearch, self).__init__(*args, **kwargs)

        self.db = SearchDatabase(self.shared)

        self.actions.add_complex_action('search-new', self.action_search_new_cb)
        self.actions.add_complex_action('search-edit', self.action_search_edit_cb)
        self.actions.add_complex_action('search-delete', self.action_search_delete_cb)

        self.left_store = Gtk.ListStore()
        self.left_store.set_column_types([str])
        self.left_treeview.set_model(self.left_store)
        self.left_treeview.append_column(Gtk.TreeViewColumn(None, Gtk.CellRendererPixbuf(icon_name='folder-saved-search-symbolic')))
        self.left_treeview.append_column(Gtk.TreeViewColumn(_('Saved search name'), Gtk.CellRendererText(), text=0))
        self.left_treeview.connect('row-activated', self.left_treeview_row_activated_cb)

        self.current_search = None
        self.setup_searches()

        self.config.setdefault('edit_dialog_size', [500,500])
        new_name = _('<New saved search name>')
        self.ssde_struct = ssde.Dict(
            label=_('Saved search'),
            substructs=[
                ssde.Text(label=_('Saved search name'), name='name', default=new_name, validator=lambda x: x != new_name),
                ssde.List(
                    label=_('Search for one of the following'), name='search',
                    substruct=ssde.TupleAndList(
                        label=_('Search for all of the following'),
                        substructs=[
                            ssde.Boolean(label=_('Exact search')),
                            ssde.Tuple(
                                label=_('Search criterion'),
                                substructs=[
                                    ssde.Choice(label=_('Field'), choices=self.fields.basic_names),
                                    ssde.Text(label=_('Value'), validator=lambda x: x),
                                ])])),
                ssde.List(
                    label=_('Sort by'), name='sort',
                    substruct=ssde.Tuple(
                        substructs=[
                            ssde.Choice(label=_('Field'), choices=self.fields.names),
                            ssde.Boolean(label=_('Reversed'))
                        ])),
                ssde.Integer(label=_('Separator depth'), name='separator'),
            ])

    @ampd.worker
    def connect_cb(self):
        while True:
            self.set_search(force=True)
            yield ampd.DATABASE

    def left_treeview_selection_changed_cb(self, left_treeview):
        try:
            self.set_search()
        except:
            pass

    def setup_searches(self, target_search=None):
        search_names = self.db.get_search_names()
        self.left_store_set_rows(search_names)
        if target_search:
            i = search_names.index(target_search)
            self.left_treeview.set_cursor(Gtk.TreePath.new_from_indices([i]))

    def action_search_new_cb(self, *args):
        value = self.ssde_struct.dialog(self.win, size=self.config['edit_dialog_size'], scrolled=True)
        if value:
            self.db.add_search(value)
            self.setup_searches(value['name'])

    def action_search_edit_cb(self, *args):
        path, column = self.left_treeview.get_cursor()
        if path:
            self.edit_search(path)

    def edit_search(self, p):
        name = self.left_store.get_value(self.left_store.get_iter(p), 0)
        value = self.db.get_search(name)
        new_value = self.ssde_struct.dialog(self.win, value, self.config['edit_dialog_size'], scrolled=True)
        if new_value:
            self.db.delete_search(name)
            self.db.add_search(value)
            self.current_search = None
            self.setup_searches(new_value['name'])

    def action_search_delete_cb(self, *args):
        path, column = self.left_treeview.get_cursor()
        if not path:
            return
        name = self.left_store.get_value(self.left_store.get_iter(path), 0)

        dialog = Gtk.Dialog(parent=self.win, title=_('Delete search'))
        dialog.add_button(_('_Cancel'), Gtk.ResponseType.CANCEL)
        dialog.add_button(_('_OK'), Gtk.ResponseType.OK)
        dialog.get_content_area().add(Gtk.Label(label=_('Delete search "{}"?').format(name), visible=True))
        reply = dialog.run()
        dialog.destroy()
        if reply == Gtk.ResponseType.OK:
            self.db.delete_search(name)
            self.setup_searches()
            self.set_search()

    def left_treeview_row_activated_cb(self, left_treeview, p, col):
        self.edit_search(p)

    @ampd.worker
    def set_search(self, force=False):
        store, i = self.left_treeview.get_selection().get_selected()
        if not i:
            self.set_songs([])
            return
        name = self.left_store.get_value(i, 0)
        if name == self.current_search and not force:
            return
        search = self.db.get_search(name)

        songs = []
        for criterion in search['search']:
            songs += yield (ampd.find if criterion[0] else ampd.search)(*sum(criterion[1], []))
        self.fields.songs_set_fields(songs)
        sort = search['sort']
        for key, reverse in reversed(sort):
            songs.sort(key=lambda song: song.get(key, ''), reverse=reverse)
        sep = search['separator']
        for i in reversed(range(len(songs)-1)):
            for j in reversed(range(sep)):
                if songs[i][sort[j][0]] != songs[i+1][sort[j][0]]:
                    songs[i+1:i+1] = (sep-j)*[self.separator_song]
                    break
        self.set_songs(songs, False)
        self.current_search = name

    def action_reset_cb(self, *args):
        self.store.set_sort_column_id(-1, Gtk.SortType.ASCENDING)
        self.current_search = None
        self.set_search()


class SearchDatabase(object):
    def __init__(self, shared):
        super(SearchDatabase, self).__init__()
        self.shared = shared
        self.get_connection()
        self.setup_database()

    def get_connection(self):
        self.connection = self.shared.sqlite_connection(SavedSearch.name)
        self.connection.cursor().execute('PRAGMA foreign_keys=ON')

    def setup_database(self):
        self.connection.cursor().execute('CREATE TABLE IF NOT EXISTS searches(name TEXT PRIMARY KEY, search TEXT, sort TEXT, separator INTEGER)')

    def get_search_names(self):
        return (name for (name,) in self.connection.cursor().execute('SELECT name FROM searches'))

    def get_search(self, name):
        t = self.connection.cursor().execute('SELECT search,sort,separator FROM searches WHERE name=?', (name,)).fetchone()
        return {'name': name, 'search': eval(t[0]), 'sort': eval(t[1]), 'separator': t[2]}

    def add_search(self, search):
        with self.connection:
            search['search'] = repr(search['search'])
            search['sort'] = repr(search['sort'])
            self.connection.cursor().execute('INSERT OR IGNORE INTO searchs(name) VALUES :name', search)
            self.connection.cursor().execute('UPDATE searchs SET search=:search,sort=:sort,separator=:separator WHERE name=:name', search)

    def delete_search(self, name):
        self.connection.cursor().execute('DELETE FROM searchs WHERE name=?', (name,))


__components__ = [SavedSearch]
__depends__ = ['songlist']
