# coding: utf-8
#
# BMPC Music Player Client
#
# Copyright (C) 2015 Itaï BEN YAACOV
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from gi.repository import Gtk, Gdk
import ast

import ampd

from bmpc import component
from bmpc import data


class SongList(component.Component):
    sortable = False
    duplicate_test_columns = []
    context_menu_paths = []

    def __init__(self, shared):
        super(SongList, self).__init__(shared)
        self.fields = self.shared.plugins['songlist'].fields

        dndtargets = [Gtk.TargetEntry.new('BMPC_SONGS', Gtk.TargetFlags.SAME_APP, 0)]

        self.treeview = data.DictTreeView(self.fields, self.data_func, self.sortable)
        self.treeview.get_selection().set_mode(Gtk.SelectionMode.MULTIPLE)

        self.store = self.treeview.get_model()
        self.total_songs = 0
        self.total_time = 0

        for action, cb, extra_params in self.shared.plugins['songlist'].actions:
            self.actions.add_complex_action(action, activate_cb=self.generic_action_cb, data=(cb,), shared=self.shared, **extra_params)

        if self.action_reset_cb != NotImplemented:
            self.actions.add_complex_action('reset', self.action_reset_cb)
        self.actions.add_complex_action('copy', self.action_copy_delete_cb)

        if self.new_song != NotImplemented:
            self.actions.add_complex_action('paste', self.action_paste_cb)
            self.actions.add_complex_action('paste-before', self.action_paste_cb)
            self.treeview.enable_model_drag_dest(dndtargets, Gdk.DragAction.MOVE | Gdk.DragAction.COPY)
            self.treeview.connect('drag-data-received', self.drag_data_received_cb)

        if self.delete_song != NotImplemented:
            self.actions.add_complex_action('delete', self.action_copy_delete_cb)
            self.actions.add_complex_action('cut', self.action_copy_delete_cb)
            self.treeview.drag_source_set(Gdk.ModifierType.BUTTON1_MASK, dndtargets, Gdk.DragAction.MOVE | Gdk.DragAction.COPY)
            self.treeview.connect('drag-data-delete', self.drag_data_delete_cb)
        else:
            self.treeview.drag_source_set(Gdk.ModifierType.BUTTON1_MASK, dndtargets, Gdk.DragAction.COPY)

        self.treeview.connect('drag-begin', self.drag_begin_cb)
        self.treeview.connect('drag-data-get', self.drag_data_get_cb)

        self.widget = self.scrolled_treeview = Gtk.ScrolledWindow(visible=True)
        self.widget.add(self.treeview)

        self.setup_context_menu(self.context_menu_paths + self.shared.plugins['songlist'].context_menu_paths, self.treeview)
        self.treeview.connect('key-press-event', self.treeview_key_press_event_cb)
        self.treeview.connect('row-activated', self.treeview_row_activated_cb)

    def close(self):
        self.treeview.destroy()
        super(SongList, self).close()

    def start(self, win):
        super(SongList, self).start(win)
        self.update_label_component()
        self.color = self.win.get_style_context().get_background_color(Gtk.StateFlags.NORMAL)

    def action_copy_delete_cb(self, action, parameter):
        songs, rows = self.treeview.get_selection_rows()
        if action.get_name() in ['copy', 'cut']:
            Gtk.Clipboard.get(Gdk.SELECTION_CLIPBOARD).set_text(repr(songs), -1)
        if action.get_name() in ['delete', 'cut']:
            self.delete_rows(rows)

    def delete_rows(self, rows):
        for row in rows:
            i = self.store.get_iter(row.get_path())
            if self.delete_song(i):
                self.store.remove(i)

    def action_paste_cb(self, action, parameter):
        Gtk.Clipboard.get(Gdk.SELECTION_CLIPBOARD).request_text(self.clipboard_paste_cb, action.get_name().endswith('before'))

    def clipboard_paste_cb(self, clipboard, raw, before):
        path, column = self.treeview.get_cursor()
        try:
            songs = ast.literal_eval(raw)
        except:
            return
        if not isinstance(songs, list) and all(isinstance(song, dict) for song in songs):
            return
        self.paste_at(songs, path, before)

    def paste_at(self, songs, path, before):
        selection = self.treeview.get_selection()
        selection.unselect_all()
        i = self.store.get_iter(path) if path else None
        if i and before:
            i = self.store.iter_previous(i)
        cursor_set = False
        for song in songs:
            if not i:
                j = self.store.insert(0)
            else:
                j = self.store.insert_after(i)
            self.store.set_row(j, song)
            if self.new_song(j):
                i = j
                if not cursor_set:
                    cursor_set = True
                    self.treeview.set_cursor(self.store.get_path(i))
                selection.select_iter(i)
            else:
                self.store.remove(j)

    def treeview_key_press_event_cb(self, treeview, event):
        if event.keyval == Gdk.KEY_Delete and event.state & (Gdk.ModifierType.SHIFT_MASK | Gdk.ModifierType.CONTROL_MASK | Gdk.ModifierType.MOD1_MASK) == 0:
            self.actions.lookup('delete').activate()
            return True
        return False

    @ampd.worker
    def treeview_row_activated_cb(self, treeview, p, column):
        if self.shared.safe_mode:
            return
        filename = self.store.get_row(self.store.get_iter(p)).file
        songs = yield ampd.playlistfind('file', filename)
        if songs:
            song_id = sorted(songs, key=lambda song: song['pos'])[0]['id']
        else:
            song_id = yield ampd.addid(filename)
        yield ampd.playid(song_id)

    def get_filenames(self, selection=True):
        if selection:
            store, paths = self.treeview.get_selection().get_selected_rows()
            rows = (store.get_row(store.get_iter(p)) for p in paths)
        else:
            rows = (self.store.get_row(self.store.iter_nth_child(None, i)) for i in range(self.store.iter_n_children()))
        return (row.file for row in rows if row)

    def generic_action_cb(self, action, parameter, cb):
        cb(self, action, parameter)

    def data_func(self, column, renderer, store, i, j):
        mix_colors = lambda r, g, b: Gdk.RGBA(self.color.red * r + 0.5 * (1-r), self.color.green * g + 0.5 * (1-g), self.color.blue * b + 0.5 * (1-b), 1.0)

        if store.get_path(i).get_indices()[0] % 2:
            renderer.set_property('background-rgba', mix_colors(0.8, 0.8, 0.8))
        else:
            renderer.set_property('background-rgba', self.color)

        _duplicate = store.get_row(i)._duplicate
        if _duplicate:
            N = 4
            _duplicate = int(_duplicate) % (N ** 3 - 1)
            f = lambda k : float((_duplicate // (N ** k)) % N) / (N-1)
            renderer.set_property('background-rgba', mix_colors(f(2), f(1), f(0)))

    def set_songs(self, songs, set_fields=True):
        if set_fields:
            self.fields.songs_set_fields(songs)
        if self.duplicate_test_columns:
            self.find_duplicates(songs, self.duplicate_test_columns)
        self.store.set_rows(songs)
        self.update_totals()

    @staticmethod
    def find_duplicates(songs, test_fields):
        dup_marker = 0
        dup_dict = {}
        for song in songs:
            if 'time' not in song or int(song['time']) <= 3:
                continue
            test = tuple(song.get(field) for field in test_fields)
            duplicates = dup_dict.get(test)
            if duplicates:
                if len(duplicates) == 1:
                    duplicates[0]['_duplicate'] = str(dup_marker)
                    dup_marker += 1
                song['_duplicate'] = duplicates[0]['_duplicate']
                duplicates.append(song)
            else:
                dup_dict[test] = [song]
                song.pop('_duplicate', None)

    def update_totals(self):
        self.total_songs = self.store.iter_n_children(None)
        self.total_time = 0
        i = self.store.get_iter_first()
        while i:
            t = self.store.get_row(i).time
            if t:
                self.total_time += int(t)
            i = self.store.iter_next(i)
        self.update_label_component()

    def update_label_component(self):
        if self.win:
            self.win.label_component.set_label('[{} / {}]'.format(self.total_songs, self.shared.format_time(self.total_time)))
            self.win.label_component.set_visible(True)

    def drag_begin_cb(self, treeview, context):
        context.drag_songs, context.drag_rows = self.treeview.get_selection_rows()

    def drag_data_get_cb(self, treeview, context, data, info, time):
        data.set(data.get_target(), 8, repr(context.drag_songs).encode())

    def drag_data_delete_cb(self, treeview, context):
        self.delete_rows(context.drag_rows)

    def drag_data_received_cb(self, treeview, context, x, y, data, info, time):
        temp = self.treeview.get_dest_row_at_pos(x, y)
        dest_path, position = temp or (Gtk.TreePath.new_from_indices([self.store.iter_n_children(None)-1]), Gtk.TreeViewDropPosition.AFTER)
        songs = ast.literal_eval(data.get_data().decode())
        self.paste_at(songs, dest_path, position in [Gtk.TreeViewDropPosition.BEFORE, Gtk.TreeViewDropPosition.INTO_OR_BEFORE])
        Gtk.drag_finish(context, True, context.get_selected_action() == Gdk.DragAction.MOVE, time)

    action_reset_cb = NotImplemented
    delete_song = new_song = NotImplemented


class PanedComponent(component.Component):
    left_context_menu_paths = []

    def __init__(self, shared):
        super(PanedComponent, self).__init__(shared)

        self.left_treeview = Gtk.TreeView(visible=True)
        self.scrolled_left_treeview = Gtk.ScrolledWindow(visible=True)
        self.scrolled_left_treeview.add(self.left_treeview)
        self.left_treeview.get_selection().connect('changed', self.left_treeview_selection_changed_cb)
        self.left_treeview.set_search_equal_func(lambda store, col, key, i: key.lower() not in store.get_value(i, col).lower())

        self.paned = Gtk.Paned(orientation=Gtk.Orientation.HORIZONTAL, position=self.config.setdefault('pane_separator', 100), visible=True)
        self.paned.connect('notify::position', self.paned_notify_position_cb)
        self.paned.add1(self.scrolled_left_treeview)
        self.paned.add2(self.widget)
        self.widget = self.paned

        self.setup_context_menu(self.left_context_menu_paths + shared.plugins['songlist'].left_context_menu_paths, self.left_treeview)

        self.starting = True

    def start(self, win):
        if self.starting:
            self.left_treeview.grab_focus()
            self.starting = False
        super(PanedComponent, self).start(win)

    def paned_notify_position_cb(self, *args):
        self.config['pane_separator'] = self.paned.get_position()

    def left_store_set_rows(self, rows):
        data.store_set_rows(self.left_store, rows, lambda i, name: self.left_store.set_value(i, 0, name))


class EditableSongList(SongList):
    SONG_NEW = 1
    SONG_DELETED = 2

    def data_func(self, column, renderer, store, i, j):
        super(EditableSongList, self).data_func(column, renderer, store, i, j)
        even = store.get_path(i).get_indices()[0] % 2 == 0
        _status = store.get_row(i)._status
        if _status == self.SONG_DELETED:
            renderer.set_property('background-rgba', Gdk.RGBA(1.0 if even else 0.9, 0.0, 0.0, 1.0))
            renderer.set_property('font', 'italic')
            renderer.set_property('strikethrough', True)
        elif _status == self.SONG_NEW:
            renderer.set_property('background-rgba', Gdk.RGBA(0.0, 1.0 if even else 0.9, 0.0, 1.0))
            renderer.set_property('font', 'bold')
            renderer.set_property('strikethrough', False)
        else:
            renderer.set_property('font', None)
            renderer.set_property('strikethrough', False)

    def __init__(self, shared):
        super(EditableSongList, self).__init__(shared)

        self.actions.add_complex_action('save', self.action_save_cb)
        self.actions.add_complex_action('undelete', self.action_undelete_cb)

    def action_undelete_cb(self, action, parameter):
        store, paths = self.treeview.get_selection().get_selected_rows()
        for p in paths:
            i = self.store.get_iter(p)
            if self.store.get_row(i)._status == self.SONG_DELETED:
                del self.store.get_row(i)._status
        self.treeview.queue_draw()

    def treeview_key_press_event_cb(self, treeview, event):
        if event.keyval == Gdk.KEY_Delete and event.state & (Gdk.ModifierType.SHIFT_MASK | Gdk.ModifierType.CONTROL_MASK | Gdk.ModifierType.MOD1_MASK) == Gdk.ModifierType.MOD1_MASK:
            self.actions.lookup('undelete').activate()
            return True
        return super(EditableSongList, self).treeview_key_press_event_cb(treeview, event)

    def new_song(self, i):
        self.set_modified()
        self.store.get_row(i)._status = self.SONG_NEW
        self.merge_new_del(i)
        return True

    def delete_song(self, i):
        self.set_modified()
        if self.store.get_row(i)._status == self.SONG_NEW:
            return True
        else:
            self.store.get_row(i)._status = self.SONG_DELETED
            self.merge_new_del(i)
            return False

    def merge_new_del(self, i):
        _status = self.store.get_row(i)._status
        for f in [self.store.iter_previous, self.store.iter_next]:
            j = f(i)
            if j and self.store.get_row(j).file == self.store.get_row(i).file and {_status, self.store.get_row(j)._status} == {self.SONG_DELETED, self.SONG_NEW}:
                del self.store.get_row(j)._status
                self.store.remove(i)
                return

    def action_copy_delete_cb(self, action, parameter):
        super(EditableSongList, self).action_copy_delete_cb(action, parameter)
        if action.get_name() in ['delete', 'cut']:
            self.treeview.queue_draw()


class SongListWithSeparator(SongList):
    SEPARATOR_FILE = 'silence3.ogg'

    def __init__(self, shared):
        super(SongListWithSeparator, self).__init__(shared)
        self.separator_song = {'file': self.SEPARATOR_FILE}
        self.get_separator_song()

    @ampd.worker
    def get_separator_song(self):
        yield ampd.CONNECT
        song = yield ampd.find('file', self.SEPARATOR_FILE)
        self.separator_song.update(song[0])
        self.treeview.queue_draw()


__menubar_paths__ = [('20#edit/10#global/10', 'comp.save', _('Save'), ['<Control>s']),
                     ('20#edit/10#global/20', 'comp.reset', _('Reset'), ['<Control>r'])]

context_menu_paths = []
for item in [('10', 'comp.cut', _('Cut'), ['<Control>x']),
             ('20', 'comp.copy', _('Copy'), ['<Control>c']),
             ('30', 'comp.paste', _('Paste'), ['<Control>v']),
             ('40', 'comp.paste-before', _('Paste before'), ['<Control>b'])]:
    __menubar_paths__.append(['20#edit/20#local/'+item[0]] + list(item[1:]))
    context_menu_paths.append(['50#edit/'+item[0]] + list(item[1:]))
for item in [('50', 'comp.delete', _('Delete'), ['Delete']),
             ('60', 'comp.undelete', _('Undelete'), ['<Alt>Delete'])]:
    context_menu_paths.append(['50#edit/'+item[0]] + list(item[1:]))


def song_title(song):
    title = song.get('title') or song.get('name', '')
    filename = song.get('file', '')
    if filename.startswith('http://'):
        http_title = filename[filename.rfind('/')+1:]
        http_title = http_title[:http_title.rfind('.')]
        title = '{0} [{1}]'.format(title, http_title) if title else http_title
    return title


def __register__(plugin_data, shared):
    plugin_data.actions = []
    plugin_data.context_menu_paths = context_menu_paths
    plugin_data.left_context_menu_paths = []
    plugin_data.fields = data.FieldFamily(shared.config.setdefault('songlist', {}).setdefault('fields', {}))
    plugin_data.fields.register_field(data.Field('album', _('Album')))
    plugin_data.fields.register_field(data.Field('albumartist', _('Album artist')))
    plugin_data.fields.register_field(data.Field('artist', _('Artist')))
    plugin_data.fields.register_field(data.Field('composer', _('Composer')))
    plugin_data.fields.register_field(data.Field('date', _('Date')))
    plugin_data.fields.register_field(data.Field('disc', _('Disc')))
    plugin_data.fields.register_field(data.Field('file', _('File')))
    plugin_data.fields.register_field(data.Field('genre', _('Genre')))
    #plugin_data.fields.register_field(data.Field('last-modified', _('Last modified')))
    plugin_data.fields.register_field(data.Field('performer', _('Performer')))
    plugin_data.fields.register_field(data.Field('time', _('Seconds'), visible=False))
    plugin_data.fields.register_field(data.Field('time_fmt', _('Duration'), get_value=lambda song: shared.format_time(song['time']) if 'time' in song else ''))
    plugin_data.fields.register_field(data.Field('title', _('Title (partial)')))
    plugin_data.fields.register_field(data.Field('title_full', _('Title'), get_value=song_title))
    plugin_data.fields.register_field(data.Field('track', _('Track')))
    plugin_data.fields.register_field(data.FieldWithTable(
        'extension', _('Extension'),
        table=[
            [
                'file',
                '\\.([^.]*)$',
                '\\1'
            ]
        ]))
    plugin_data.fields.register_field(data.FieldWithTable(
        'agenre', visible=False,
        table=[
            [
                'genre',
                '[Mm]ilong',
                'b milonga'
            ],
            [
                'genre',
                '[Cc]andombe',
                'b milonga'
            ],
            [
                'genre',
                '[Tt]ango|Canci[oó]n',
                'a tango'
            ],
            [
                'genre',
                '[Vv]als',
                'c vals'
            ],
            [
                'genre',
                '[Ff]ox ?trot',
                'd fox'
            ],
            [
                'genre',
                '[Pp]aso ?doble',
                'e paso'
            ],
            [
                'genre',
                'Ranchera',
                'f ranchera'
            ],
            [
                None,
                None,
                'z'
            ]
        ]))
    plugin_data.fields.register_field(data.FieldWithTable(
        'artist_sort_name', visible=False,
        table=[
            [
                'artist',
                '(La Típica Sanata|Otros Aires)',
                '\\1'
            ],
            [
                'artist',
                '^(.* Tango)$',
                '\\1'
            ],
            [
                'artist',
                '(.*), dir\. (.*) ([^ ]+)',
                '\\3, \\2 (\\1)'
            ],
            [
                'artist',
                '(Orquesta Típica|Dúo|Cuarteto|Sexteto) (.*)',
                '\\2, \\1'
            ],
            [
                'artist',
                '(.*) ((?:Di|De) *[^ ]+)',
                '\\2, \\1'
            ],
            [
                'artist',
                '(.*) ([^ ]+)',
                '\\2, \\1'
            ],
            [
                'artist',
                '(.*)',
                '\\1'
            ]
        ]))
    performer_last_name = data.FieldWithTable(
        'performer_last_name', visible=False,
        table=[
            [
                'performer',
                '^(.*) ((?:Di|De|Del) *[^ ]+)$',
                '\\2'
            ],
            [
                'performer',
                '^(.*) ([^ ]+)$',
                '\\2'
            ],
            [
                'performer',
                '^(.*)$',
                '\\1'
            ]
        ])
    plugin_data.fields.register_field(data.Field(
        'performers_last_names', visible=False,
        get_value=lambda song: ', '.join(performer_last_name.get_value({'performer': name}) for name in song.get('performer').split(', ')) if song.get('performer') else None))
