# coding: utf-8
#
# BMPC Music Player Client
#
# Copyright (C) 2015 Itaï BEN YAACOV
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from gi.repository import GObject, Gtk, Gdk

import re
import datetime
import ast

import ampd

from bmpc import data
from bmpc.utils.logger import logger

from . import songlist


class Tanda(songlist.PanedComponent):
    title = _('Tanda Editor')
    name = 'tanda'
    key = '6'
    action_prefix = 'supercomp'

    GENRES = ('Tango', 'Vals', 'Milonga', _('Other'), _('All'))
    GENRE_OTHER = len(GENRES) - 2
    GENRE_ALL = len(GENRES) - 1

    def __init__(self, shared):
        self.widget = self.right_box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, visible=True)
        super(Tanda, self).__init__(shared)

        self.db = self.shared.plugins[self.name].db
        self.actions.add_complex_action('tanda-toggle-view', self.action_toggle_view_cb)
        self.actions.add_complex_action('tanda-verify-db', self.db.action_tanda_verify_db_cb)
        self.actions.add_complex_action('tanda-reconnect', self.db.action_reconnect_cb)

        self.all_tandas = self.filtered_tandas = []

        self.left_store = Gtk.ListStore(str)
        self.left_treeview.set_model(self.left_store)
        self.left_treeview.insert_column_with_attributes(0, _('Artist'), Gtk.CellRendererText(), text=0)
        self.left_treeview.get_selection().set_mode(Gtk.SelectionMode.MULTIPLE)

        self.button_box = Gtk.ButtonBox(orientation=Gtk.Orientation.HORIZONTAL, layout_style=Gtk.ButtonBoxStyle.START, visible=True)
        button = None
        for i, genre in enumerate(self.GENRES):
            button = Gtk.RadioButton(label=genre, visible=True, can_focus=False, group=button)
            button.connect('toggled', self.genre_button_toggled_cb, i)
            button.set_mode(False)
            self.button_box.pack_start(button, False, False, 0)
        self.genre_filter = 0

        self.problem_button = Gtk.ToggleButton(image=Gtk.Image(icon_name='dialog-warning-symbolic'), visible=True, can_focus=False, active=self.shared.safe_mode)
        self.problem_button.connect('toggled', self.problem_button_toggled_cb)
        self.view_button = Gtk.ToggleButton(image=Gtk.Image(icon_name='view-list-symbolic'), visible=True, can_focus=False)
        self.view_button.connect('toggled', self.view_button_toggled_cb)
        self.button_box.pack_end(self.problem_button, False, False, 0)
        self.button_box.set_child_secondary(self.problem_button, True)
        self.button_box.pack_end(self.view_button, False, False, 0)
        self.button_box.set_child_secondary(self.view_button, True)
        self.safe_mode_handler = self.shared.connect('notify::safe-mode', lambda shared, param_spec: shared.safe_mode and self.problem_button.set_active(True))

        self.right_box.pack_start(self.button_box, False, False, 0)

        self.db.connect('changed', self.db_changed_cb)

        self.sub_component = self.edit = TandaEdit(shared)
        self.view = TandaView(shared)
        self.read_db()

    def close(self):
        super(Tanda, self).close()
        self.db.disconnect_by_func(self.db_changed_cb)
        self.edit.close()
        self.view.close()
        self.shared.handler_disconnect(self.safe_mode_handler)

    def start(self, win):
        super(Tanda, self).start(win)
        self.sub_component.start(win)
        self.right_box.add(self.sub_component.widget)

    def stop(self):
        super(Tanda, self).stop()
        self.sub_component.stop()
        self.right_box.remove(self.sub_component.widget)

    def action_toggle_view_cb(self, action, param):
        self.view_button.activate()

    def genre_button_toggled_cb(self, button, i):
        if button.get_active():
            self.genre_filter = i
            self.filter_tandas(False)

    def problem_button_toggled_cb(self, button):
        self.filter_tandas(False)

    def view_button_toggled_cb(self, button):
        current_tandaid = self.sub_component.current_tandaid
        self.sub_component.stop()
        self.right_box.remove(self.sub_component.widget)
        self.sub_component = self.view if button.get_active() else self.edit
        self.sub_component.start(self.win)
        self.right_box.add(self.sub_component.widget)
        self.left_treeview_selection_changed_cb()
        self.sub_component.set_cursor_tandaid(current_tandaid)

    def tanda_filter_holds(self, tanda):
        if 'problems' in tanda and self.problem_button.get_active():
            return False
        if self.genre_filter == self.GENRE_ALL:
            return True
        for i in range(self.GENRE_OTHER):
            if self.GENRES[i] in tanda.get('genre', ''):
                return self.genre_filter == i
        return self.genre_filter == self.GENRE_OTHER

    def read_db(self):
        self.all_tandas = list(self.db.get_tandas())
        self.filter_tandas()

    def filter_tandas(self, sort=True):
        if sort:
            self.all_tandas.sort(key=self.tanda_key_func)
        self.filtered_tandas = list(filter(self.tanda_filter_holds, self.all_tandas))
        artists = sorted(set(tanda.get('artist', '') for tanda in self.all_tandas))
        self.left_store_set_rows(['*'] + artists)
        self.left_treeview_selection_changed_cb()

    def left_treeview_selection_changed_cb(self, *args):
        if not self.filtered_tandas:
            return
        store, paths = self.left_treeview.get_selection().get_selected_rows()
        artists = [store.get_value(store.get_iter(path), 0) for path in paths] or ['*']
        current_tandaid = self.sub_component.current_tandaid
        self.sub_component.set_tandas(filter(lambda tanda: '*' in artists or tanda.get('artist') in artists, self.filtered_tandas))
        self.sub_component.set_cursor_tandaid(current_tandaid)

    def db_changed_cb(self, db, tandaid):
        if tandaid == -1:
            return self.read_db()

        for tanda in self.all_tandas:
            if tanda.get('tandaid') == tandaid:
                if not self.sub_component.reread_tanda(tanda):
                    self.all_tandas.remove(tanda)
                break
        else:
            tanda = self.db.get_tanda(tandaid)
            if tanda:
                self.all_tandas.append(tanda)
        self.filter_tandas()

    @staticmethod
    def tanda_key_func(tanda):
        return (tanda.get('artist', ''),
                99 if 'genre' not in tanda else 1 if 'Tango' in tanda['genre'] else 2 if 'Vals' in tanda['genre'] else 3 if 'Milonga' in tanda['genre'] else 4,
                tanda['years'][2:] if 'years' in tanda and tanda['years'].startswith('* ') else tanda.get('years', ''),
                tanda.get('performer', ''),
                tanda.get('first_song', ''),)


class TandaSubComponent(object):
    name = Tanda.name # For config

    def __init__(self, shared):
        super(TandaSubComponent, self).__init__(shared)
        self.db = self.shared.plugins[self.name].db
        self.current_tandaid = None

    def init_tandaid_treeview(self, treeview):
        self.tandaid_treeview = treeview
        self.tandaid_store = treeview.get_model()
        self.tandaid_treeview.connect('cursor-changed', self.tandaid_cursor_changed_cb)

    def reread_tanda(self, tanda):
        db_tanda = self.db.get_tanda(tanda['tandaid'])
        if db_tanda == None:
            return False
        for name in list(tanda.keys()):
            if not name in db_tanda:
                del tanda[name]
        tanda.update(db_tanda)
        return True

    def set_cursor_tandaid(self, tandaid):
        for i in self.tandaid_store.iters():
            if self.tandaid_store.get_row(i).tandaid == tandaid:
                path = self.tandaid_store.get_path(i)
                self.tandaid_treeview.set_cursor(self.tandaid_store.get_path(i))
                self.tandaid_treeview.scroll_to_cell(path, None, True, 0.5, 0.0)
                return

    def tandaid_cursor_changed_cb(self, treeview):
        path, column = self.tandaid_treeview.get_cursor()
        self.current_tandaid = path and self.tandaid_store.get_row(self.tandaid_store.get_iter(path)).tandaid


class TandaEdit(TandaSubComponent, songlist.EditableSongList):
    def __init__(self, shared):
        super(TandaEdit, self).__init__(shared)

        self.current_tanda = None
        self.current_tanda_path = None

        self.actions.add_complex_action('tanda-delete', self.action_tanda_delete_cb)
        self.actions.add_complex_action('tanda-reset-field', self.action_tanda_field_cb)
        self.actions.add_complex_action('tanda-fill-field', self.action_tanda_field_cb)
        self.actions.add_complex_action('tanda-merge', self.action_tanda_merge_cb)  #XXXXXXXXXXXXXXX

        self.tanda_treeview = data.DictTreeView(self.db.fields, self.tanda_data_func, True)
        self.tanda_treeview.get_selection().set_mode(Gtk.SelectionMode.MULTIPLE)  #XXXXXXXXXXXXXXXXXX
        css_provider = Gtk.CssProvider.new()
        css_provider.load_from_data(b'GtkTreeView.view { outline-width: 4px; outline-style: solid; }')
        self.tanda_treeview.get_style_context().add_provider(css_provider, 1000)
        self.tanda_treeview.connect('button-press-event', self.tanda_treeview_button_press_event_cb)
        self.setup_context_menu([('50/10', 'comp.tanda-delete', _('Delete tanda'))] + self.shared.plugins['songlist'].left_context_menu_paths, self.tanda_treeview)
        self.init_tandaid_treeview(self.tanda_treeview)

        for name in self.db.fields.basic_names:
            col = self.tanda_treeview.cols[name]
            col.renderer.set_property('editable', True)
            col.renderer.connect('editing-started', self.renderer_editing_started_cb, name)
        self.tanda_store = self.tanda_treeview.get_model()
        self.tanda_treeview.connect('cursor-changed', self.tanda_treeview_cursor_changed_cb)
        scrolled_tanda_treeview = Gtk.ScrolledWindow(visible=True)
        scrolled_tanda_treeview.add(self.tanda_treeview)

        self.treeview.set_vexpand(False)
        self.scrolled_treeview.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.NEVER)

        self.widget = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, visible=True)
        self.widget.pack_start(scrolled_tanda_treeview, True, True, 0)
        self.widget.pack_start(self.scrolled_treeview, False, False, 0)

        self.queue = []

        self.redraw_worker = None
        self.scrolled_treeview.connect('size-allocate', self.size_allocate_cb)

    @ampd.worker
    def connect_cb(self):
        while True:
            self.queue = yield ampd.playlistinfo()
            if self.current_tanda:
                self.find_duplicates(self.queue + self.current_tanda._songs, ['title'])
                self.treeview.queue_draw()
            yield ampd.PLAYLIST

    @staticmethod
    def tanda_data_func(column, renderer, store, i, j):
        tanda = store.get_row(i)
        if tanda._modified:
            renderer.set_property('font', 'bold italic')
        else:
            renderer.set_property('font', None)

        if 'last_played' in column.field.name and tanda.last_played_weeks != None:
            t = min(tanda.last_played_weeks / 20.0, 1.0)
            renderer.set_property('background-rgba', Gdk.RGBA(1.0-t,t,0.0,1.0))
            renderer.set_property('background-set', True)
        elif column.field.name in ('drama', 'energy', 'romance', 'rhythm', 'speed') and column.field.name in tanda.get_data():
            t = min((int(tanda.get_data()[column.field.name])-1) / 4.0, 1.0)
            renderer.set_property('background-rgba', Gdk.RGBA(t,0.5,1.0-t,1.0))
            renderer.set_property('background-set', True)
        else:
            renderer.set_property('background-set', False)

    def set_modified(self, modified=True):
        self.current_tanda._modified = modified
        self.tanda_treeview.queue_draw()

    def set_tandas(self, tandas):
        self.tanda_store.set_rows(tandas)
        self.set_current_tanda()

    def tanda_treeview_cursor_changed_cb(self, *args):
        self.current_tanda_path, column = self.tanda_treeview.get_cursor()
        self.set_current_tanda()

    def set_current_tanda(self):
        if self.current_tanda:
            self.current_tanda._songs = [self.store.get_row(i).get_data() for i in self.store.iters()]
        if self.current_tanda_path == None:
            self.current_tanda = None
            self.set_songs([])
        else:
            self.current_tanda = self.tanda_store.get_row(self.tanda_store.get_iter(self.current_tanda_path))
            self.find_duplicates(self.queue + self.current_tanda._songs, ['title'])
            self.set_songs(self.current_tanda._songs)
        if self.redraw_worker:
            self.redraw_worker.kill()
        self.redraw_worker = self.redraw_treeview()

    @ampd.worker
    def redraw_treeview(self):
        yield ampd.TIMEOUT(50)
        self.widget.remove(self.scrolled_treeview)
        self.widget.add(self.scrolled_treeview)
        self.redraw_worker = None

    def size_allocate_cb(self, widget, allocation):
        if self.current_tanda_path:
            self.tanda_treeview.scroll_to_cell(self.current_tanda_path, None, False)

    def renderer_editing_started_cb(self, renderer, editable, path, name):
        editable.connect('editing-done', self.editing_done_cb, path, name)
        self.shared.menubar_accels_enable(False)

    def editing_done_cb(self, editable, path, name):
        self.shared.menubar_accels_enable(True)
        if editable.get_property('editing-canceled'):
            return
        value = editable.get_text()
        if value != self.current_tanda.get_data().get(name, ''):
            if value:
                self.current_tanda.__setattr__(name, value)
            else:
                self.current_tanda.__delattr__(name)
            self.set_modified()

    def tanda_treeview_button_press_event_cb(self, tanda_treeview, event):
        pos = self.tanda_treeview.get_path_at_pos(event.x, event.y)
        if not pos:
            return False
        if event.type == Gdk.EventType.BUTTON_PRESS and event.button == 1:
            path, col, cx, xy = pos
            selection = self.tanda_treeview.get_selection()
            if event.state & Gdk.ModifierType.CONTROL_MASK:
                if selection.path_is_selected(path):
                    selection.unselect_path(path)
                else:
                    selection.select_path(path)
            elif event.state & Gdk.ModifierType.SHIFT_MASK:
                oldpath, column = self.tanda_treeview.get_cursor()
                if oldpath:
                    selection.select_range(oldpath, path)
                else:
                    selection.select_path(path)
            else:
                self.tanda_treeview.set_cursor_on_cell(path, col, col.renderer, False)
                self.tanda_treeview.grab_focus()
            return True
        elif event.type == Gdk.EventType._2BUTTON_PRESS and event.button == 1:
            path, col, cx, xy = pos
            self.tanda_treeview.set_cursor(path, col, True)
            return True

    def action_tanda_delete_cb(self, *args):
        path, column = self.tanda_treeview.get_cursor()
        if not path:
            return
        i = self.tanda_store.get_iter(path)
        tanda = self.tanda_store.get_row(i)
        title = ' / '.join(filter(lambda x: x, (tanda.artist, tanda.years, tanda.performer)))
        dialog = Gtk.Dialog(parent=self.win, title=_('Delete tanda'))
        dialog.add_button(_('_Cancel'), Gtk.ResponseType.CANCEL)
        dialog.add_button(_('_OK'), Gtk.ResponseType.OK)
        dialog.get_content_area().add(Gtk.Label(label=_('Delete {}?').format(title), visible=True))
        reply = dialog.run()
        dialog.destroy()
        if reply == Gtk.ResponseType.OK:
            self.db.delete_tanda(tanda.tandaid)

    def get_filenames(self, selection=True):
        if selection:
            return super(TandaEdit, self).get_filenames(True)
        else:
            store, paths = self.tanda_treeview.get_selection().get_selected_rows()
            return sum(([song['file'] for song in store.get_row(store.get_iter(path))._songs] + [songlist.SongListWithSeparator.SEPARATOR_FILE] for path in paths), [songlist.SongListWithSeparator.SEPARATOR_FILE])

    def action_save_cb(self, action, parameter):
        if self.current_tanda:
            self.current_tanda._songs = [self.store.get_row(i).get_data() for i in self.store.iters()]
        store, paths = self.tanda_treeview.get_selection().get_selected_rows()
        for path in paths:
            tanda = store.get_row(store.get_iter(path))
            tanda._songs = [song for song in tanda._songs if song.get('_status') != self.SONG_DELETED]
            self.db.update_tanda(tanda.get_data())

    def action_reset_cb(self, action, parameter):
        self.reread_tanda(self.current_tanda.get_data())

    def action_tanda_field_cb(self, action, parameter):
        path, column = self.tanda_treeview.get_cursor()
        if not path:
            return
        reset = 'reset' in action.get_name()
        tanda = self.tanda_store.get_row(self.tanda_store.get_iter(path))
        name = column.field.name
        alt_tanda = self.db.get_tanda(tanda.tandaid) if reset else self.db.tanda_from_songs([song for song in tanda._songs if song.get('_status') != self.SONG_DELETED])
        if name in alt_tanda:
            tanda.get_data()[name] = alt_tanda[name]
        elif reset:
            try:
                del tanda.get_data()[name]
            except:
                pass
        if not reset:
            self.set_modified()
        self.tanda_treeview.queue_draw()

    def reread_tanda(self, tanda):
        if not super(TandaEdit, self).reread_tanda(tanda):
            return False
        if self.current_tanda and tanda['tandaid'] == self.current_tanda.tandaid:
            self.set_songs(self.current_tanda._songs)
        self.tanda_treeview.queue_draw()
        return True

    def action_copy_delete_cb(self, action, parameter):
        focus = self.win.get_focus()
        if focus == self.tanda_treeview and action.get_name() == 'copy':
            path, column = focus.get_cursor()
            if column:
                name = column.field.name
                Gtk.Clipboard.get(Gdk.SELECTION_CLIPBOARD).set_text(repr({name: self.tanda_store.get_row(self.tanda_store.get_iter(path)).get_data()[name]}), -1)
        elif isinstance(focus, Gtk.Entry):
            if action.get_name() in ['copy', 'cut']:
                focus.emit(action.get_name() + '-clipboard')
            else:
                focus.emit('delete-from-cursor', Gtk.DeleteType.CHARS, 1)
        else:
            super(TandaEdit, self).action_copy_delete_cb(action, parameter)

    def action_paste_cb(self, action, parameter):
        focus = self.win.get_focus()
        if isinstance(focus, Gtk.Entry):
            focus.emit('paste-clipboard')
        else:
            super(TandaEdit, self).action_paste_cb(action, parameter)

    def clipboard_paste_cb(self, clipboard, raw, before):
        focus = self.win.get_focus()
        if focus == self.tanda_treeview:
            store, paths = focus.get_selection().get_selected_rows()
            try:
                to_paste = ast.literal_eval(raw)
            except:
                raise
            if not isinstance(to_paste, dict):
                return
            for path in paths:
                tanda = self.tanda_store.get_row(self.tanda_store.get_iter(path)).get_data()
                for name, value in to_paste.items():
                    if value:
                        tanda[name] = value
                    else:
                        tanda.pop(name, None)
                tanda['_modified'] = True
            focus.queue_draw()
        else:
            super(TandaEdit, self).clipboard_paste_cb(clipboard, raw, before)

    def action_tanda_merge_cb(self, *args):  #XXXXXXXXXXXXXXXXXX
        store, paths = self.tanda_treeview.get_selection().get_selected_rows()
        tandas = [store.get_row(store.get_iter(path)) for path in paths]
        for target in tandas:
            if all(tanda == target or tanda.problems == 'NOT FOUND' for tanda in tandas):
                break
        else:
            return
        target._songs = sum((tanda._songs for tanda in tandas), [])
        for tanda in tandas:
            tanda._modified = True
            if tanda != target:
                tanda._songs = []
            if tanda == self.current_tanda:
                self.set_songs(tanda._songs)
        self.tanda_treeview.queue_draw()
        self.set_current_tanda()


class TandaView(TandaSubComponent, songlist.SongListWithSeparator):
    duplicate_test_columns = ['file']

    def __init__(self, shared):
        super(TandaView, self).__init__(shared)
        self.init_tandaid_treeview(self.treeview)

    def set_tandas(self, tandas):
        songs = [self.separator_song.copy()]
        songs[0]['tandaid'] = None
        for tanda in tandas:
            tandaid = tanda['tandaid']
            for song in tanda['_songs'] + [self.separator_song.copy()]:
                song['tandaid'] = tandaid
                songs.append(song)
        self.set_songs(songs)


class TandaDatabase(GObject.Object):
    __gsignals__ = {
        'changed': (GObject.SIGNAL_RUN_LAST, None, (int,))
    }

    def __init__(self, shared):
        super(TandaDatabase, self).__init__()
        self.shared = shared
        self.config = shared.config.setdefault(Tanda.name, {})
        self.songlist_fields = shared.plugins['songlist'].fields
        self.ampd_worker_group = ampd.WorkerGroup(self.shared.ampd_client)

        self.fields = data.FieldFamily(self.config.setdefault('fields', {}))
        self.fields.register_field(data.Field('artist', _('Artist')))
        self.fields.register_field(data.Field('genre', _('Genre')))
        self.fields.register_field(data.Field('years_fixed', visible=False))
        self.fields.register_field(data.Field('years_min', visible=False, get_value=lambda tanda: min(song.get('date', '').split('-',1)[0] for song in tanda['_songs']) or '????' if tanda.get('_songs') else None))
        self.fields.register_field(data.Field('years_max', visible=False, get_value=lambda tanda: max(song.get('date', '').split('-',1)[0] for song in tanda['_songs']) or '????' if tanda.get('_songs') else None))
        self.fields.register_field(data.Field('years', _('Years'), get_value=lambda tanda: ('\'{}'.format(tanda['years_min'][2:]) if tanda['years_min'] == tanda['years_max'] else '\'{}-\'{}'.format(tanda['years_min'][2:], tanda['years_max'][2:])) if 'years_min' in tanda and 'years_max' in tanda else '* ' + tanda.get('years_fixed', '')))
        self.fields.register_field(data.Field('first_song', _('First song')))
        self.fields.register_field(data.Field('performer', _('Performer')))
        self.fields.register_field(data.Field('comment', _('Comment')))
        self.fields.register_field(data.Field('problems', _('Problems')))
        self.fields.register_field(data.Field('note', _('Note'), min_width=30))
        self.fields.register_field(data.Field('drama', _('Drama'), min_width=30))
        self.fields.register_field(data.Field('energy', _('Energy'), min_width=30))
        self.fields.register_field(data.Field('romance', _('Romance'), min_width=30))
        self.fields.register_field(data.Field('rhythm', _('Rhythm'), min_width=30))
        self.fields.register_field(data.Field('speed', _('Speed'), min_width=30))
        self.fields.register_field(data.Field('last_modified', _('Last modified')))
        self.fields.register_field(data.Field('last_played', _('Last played')))
        self.fields.register_field(data.Field('last_played_weeks', _('Weeks since last played'), min_width=30, get_value=lambda tanda: (datetime.date.today() - datetime.date(*map(int, tanda['last_played'].split('-')))).days // 7 if 'last_played' in tanda and tanda['last_played'] else None))
        self.fields.register_field(data.Field('n_songs', _('Number of songs'), min_width=30, get_value=lambda tanda: 0 if not tanda.get('_songs') else None if (len(tanda.get('_songs')) == 4 and tanda.get('genre') == 'Tango') or (len(tanda.get('_songs')) == 3 and tanda.get('genre') in {'Vals', 'Milonga'}) else len(tanda.get('_songs'))))

        self.get_connection()
        self.setup_database()

    def get_connection(self):
        self.connection = self.shared.sqlite_connection(Tanda.name)
        self.connection.cursor().execute('PRAGMA foreign_keys=ON')

    def setup_database(self):
        cursor = self.connection.cursor()
        cursor.execute('CREATE TABLE IF NOT EXISTS tandas(tandaid INTEGER PRIMARY KEY)')
        for field in self.fields.basic_names:
            try:
                cursor.execute('ALTER TABLE tandas ADD COLUMN ' + field)
            except:
                pass

        cursor.execute('CREATE TABLE IF NOT EXISTS songs(file TEXT NOT NULL PRIMARY KEY)')
        for name in self.songlist_fields.basic_names:
            try:
                cursor.execute('ALTER TABLE songs ADD COLUMN ' + name)
            except:
                pass
        cursor.executemany('UPDATE tandas SET genre=? WHERE genre=?', (('Tango', 'T'),
                                                                       ('Vals', 'V'),
                                                                       ('Milonga', 'M'),))

        cursor.execute('CREATE TABLE IF NOT EXISTS tanda_songs(tandaid INTEGER NOT NULL, position INTEGER NOT NULL, file TEXT NOT NULL, PRIMARY KEY(tandaid, position), FOREIGN KEY(tandaid) REFERENCES tandas, FOREIGN KEY(file) REFERENCES songs)')

    def add_song(self, song):
        cursor = self.connection.cursor()
        if cursor.execute('SELECT ? NOT IN (SELECT file FROM songs)', (song['file'],)).fetchone()[0]:
            with self.connection:
                cursor.execute('INSERT INTO songs(file) VALUES(:file)', song)
                self.update_song(song)

    def update_song(self, song):
        self.connection.cursor().execute('UPDATE songs SET {} WHERE file=:file'.format(self._make_value_list(self.songlist_fields.basic_names, exclude='file')), song)

    def set_tanda_songs(self, tandaid, songs):
        cursor = self.connection.cursor()
        tanda_songs = dict(cursor.execute('SELECT position, file FROM tanda_songs WHERE tandaid=?', (tandaid,)).fetchall())
        for position, song in enumerate(songs):
            self.add_song(song)
            if position in tanda_songs:
                if tanda_songs[position] != song['file']:
                    cursor.execute('UPDATE tanda_songs SET file=? WHERE tandaid=? AND position=?', (song['file'], tandaid, position))
            else:
                cursor.execute('INSERT INTO tanda_songs(tandaid, position, file) VALUES(?, ?, ?)', (tandaid, position, song['file']))
            if position in tanda_songs:
                del tanda_songs[position]
        cursor.executemany('DELETE FROM tanda_songs WHERE tandaid=? AND position=?', ((tandaid, position) for position in tanda_songs.keys()))

    def get_tandas(self):
        query = self.connection.cursor().execute('SELECT tandaid, {} FROM tandas'.format(','.join(self.fields.basic_names)))
        return map(self._get_tanda_from_tuple, query)

    def get_tanda(self, tandaid):
        t = self.connection.cursor().execute('SELECT tandaid, {} FROM tandas WHERE tandaid=?'.format(','.join(self.fields.basic_names)), (tandaid,)).fetchone()
        return t and self._get_tanda_from_tuple(t)

    @staticmethod
    def _tuple_to_dict(t, names):
        return {name: t[i] for i, name in enumerate(names) if t[i] != None}

    def _get_tanda_from_tuple(self, t):
        tanda = self._tuple_to_dict(t, ['tandaid'] + self.fields.basic_names)
        query = self.connection.cursor().execute('SELECT {} FROM tanda_songs,songs USING(file) WHERE tanda_songs.tandaid=?'.format(', songs.'.join(['tanda_songs.position'] + self.songlist_fields.basic_names)), (tanda['tandaid'],))
        tanda['_songs'] = [self._tuple_to_dict(t, ['_position'] + self.songlist_fields.basic_names) for t in query]
        self.fields.song_set_fields(tanda)
        return tanda

    def update_tanda(self, tanda):
        with self.connection:
            self.connection.cursor().execute('UPDATE tandas SET {} WHERE tandaid=:tandaid'.format(', '.join([name + '=:' + name for name in self.fields.basic_names if name != 'tandaid'])), tanda)
            tandaid = tanda['tandaid']
            self.set_tanda_songs(tandaid, tanda['_songs'])
        self.emit('changed', tandaid)

    def delete_tanda(self, tandaid):
        with self.connection:
            self.connection.cursor().execute('DELETE FROM tanda_songs WHERE tandaid=?; DELETE FROM tandas WHERE tandaid=?', (tandaid, tandaid))
        self.emit('changed', tandaid)

    @staticmethod
    def tanda_from_songs(songs):
        fields = (('artist_sort_name', '???'),
                  ('genre', '???'),
                  ('performers_last_names', '???'))
        merged = {field: [song.get(field, default) for song in songs] for field, default in fields}
        merged['genre'] = [genre if genre.split(' ', 1)[0] not in ('Tango', 'Vals', 'Milonga') else genre.split(' ', 1)[0] for genre in merged['genre']]
        sorted_merged = {field: sorted(set(merged[field])) for field, default in fields}
        tanda = {}
        tanda['artist'] = '; '.join(sorted_merged['artist_sort_name'])
        match = re.match(r'(.*) \((.*)\)', tanda['artist'])
        if match:
            tanda['artist'] = match.group(1)
            tanda['comment'] = match.group(2)

        performers = []
        for song_performers in merged['performers_last_names']:
            for performer in song_performers.split(', '):
                if performer not in performers:
                    performers.append(performer)
        tanda['performer'] = ', '.join(performers)
        tanda['genre'] = ', '.join(sorted_merged['genre'])
        tanda['first_song'] = songs[0]['title']
        tanda['last_modified'] = str(datetime.date.today())
        return tanda

    @staticmethod
    def _make_value_list(names, exclude=None, separator=', ', operator='='):
        return separator.join(name + operator + ':' + name for name in names if name != exclude)

    def action_tanda_define_cb(self, songlist, action, parameter):
        songs, rows = songlist.treeview.get_selection_rows()
        tanda = self.tanda_from_songs(songs)
        with self.connection:
            self.connection.cursor().execute('INSERT INTO tandas DEFAULT VALUES')
            tanda['tandaid'] = tandaid = self.connection.last_insert_rowid()
            self.connection.cursor().execute('UPDATE tandas SET {} WHERE tandaid=:tandaid'.format(self._make_value_list(self.fields.basic_names)), tanda)
            self.set_tanda_songs(tandaid, songs)
        self.emit('changed', tandaid)

    def action_reconnect_cb(self, action, param):
        self.get_connection()

    @ampd.worker
    def action_tanda_verify_db_cb(self, action, param):
        yield ampd.CONNECT
        self.get_connection() # A way to reset a database connection.
        deleted = 0
        changed = 0
        trouble = 0
        skip = []
        with self.connection:
            query = self.connection.cursor().execute('SELECT {},file in (SELECT file from tanda_songs) FROM songs'.format(', '.join(self.songlist_fields.basic_names))).fetchall()
            for song in map(lambda t: self._tuple_to_dict(t, self.songlist_fields.basic_names + ['used']), query):
                if song['file'] in skip:
                    continue
                if not song['used']:
                    logger.info(_('Deleting "{file}"').format(**song))
                    self.connection.cursor().execute('DELETE FROM songs WHERE file=:file', song)
                    deleted += 1
                    continue
                real_song = yield ampd.find('file', song['file'])
                if real_song:
                    self.update_song(real_song[0])
                    continue
                maybe_song = yield ampd.find(*sum(([field, song.get(field, '')] for field in ('artist', 'title', 'date', 'performer')), []))
                if maybe_song:
                    logger.info(_('Replacing song:'))
                    logger.info('- ' + song['file'])
                    logger.info('+ ' + maybe_song[0]['file'])
                    self.add_song(maybe_song[0])
                    self.connection.cursor().execute('UPDATE tanda_songs SET file=? WHERE file=?', (maybe_song[0]['file'], song['file']))
                    self.connection.cursor().execute('DELETE FROM songs WHERE file=:file', song)
                    changed += 1
                    skip.append(maybe_song[0]['file'])
                    continue
                logger.info(_('Cannot find "{file}"').format(**song))
                trouble += 1
            logger.info(_('Tanda database checked: {} songs unused, {} replaced, {} unknown').format(deleted, changed, trouble))
            self.emit('changed', -1)


__components__ = [Tanda]
__depends__ = ['songlist']
__menubar_paths__ = [('20#edit/90#tanda/10', 'comp.tanda-reset-field', _('Reset tanda field'), ['<Control><Shift>r']),
                     ('20#edit/90#tanda/20', 'comp.tanda-fill-field', _('Fill tanda field'), ['<Control><Shift>f']),
                     ('20#edit/90#tanda/30', 'comp.tanda-merge', _('Merge'), ['<Control><Shift>m']),
                     ('20#edit/90#tanda/40', 'comp.tanda-delete', _('Delete tanda'), ['Delete']),
                     ('20#edit/95#tanda/10', 'supercomp.tanda-toggle-view', _('Toggle tanda view mode'), ['<Control>Tab']),
                     ('20#edit/95#tanda/20', 'supercomp.tanda-verify-db', _('Verify tanda database'), ['<Control><Shift>d']),
                     ('20#edit/95#tanda/30', 'supercomp.tanda-reconnect', _('Reconnect to tanda database')),]

def __register__(plugin_data, shared):
    plugin_data.db = TandaDatabase(shared)
    songlist_shared = shared.plugins['songlist']
    songlist_shared.actions += [('tanda-define', plugin_data.db.action_tanda_define_cb, {})]
    songlist_shared.context_menu_paths.append(('60#tanda/10', 'comp.tanda-define', _('Define tanda')))
    plugin_data.actions += [
        ('tanda-verify-db', plugin_data.db.action_tanda_verify_db_cb),
    ]
