# coding: utf-8
#
# BMPC Music Player Client
#
# Copyright (C) 2015 Itaï BEN YAACOV
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from gi.repository import GObject, Gtk
import sys
import os
import xdg.BaseDirectory
import json
import apsw

import ampd

from .utils.logger import logger


class AppShared(GObject.GObject):
    current_song = GObject.property()
    volume = GObject.property(type=int)
    playtime_upper = GObject.property(type=int, default=0)
    playtime_value = GObject.property(type=int, default=0)
    want_to_connect = GObject.property(type=bool, default=False)
    state = GObject.property(type=str, default='')
    bitrate = GObject.property()
    safe_mode = GObject.property(type=bool, default=False)

    option_names = ['consume', 'random', 'repeat', 'single']
    for option in option_names:
        exec('option_{} = GObject.property(type=bool, default=False)'.format(option))

    def __init__(self):
        super(AppShared, self).__init__()
        self.uidefs = {}
        self.excepthook_orig, sys.excepthook = sys.excepthook, self.excepthook
        self.current_song = {}

        self.ampd_client = ampd.Client()
        self.ampd_host = None
        self.ampd_port = 6600
        self.ampd_worker_group = ampd.WorkerGroup(self.ampd_client)
        self.ampd_worker_group.connect_loop(self.connect_cb, self.disconnect_cb, self.connect_failed_cb)
        self.connect('notify::volume', self.notify_volume_cb)
        self.connect('notify::playtime-value', self.notify_playtime_value_cb)
        for option in self.option_names:
            self.connect('notify::option-' + option, self.notify_option_cb)

        self.plugins = {}

        self.config = {}
        for path in xdg.BaseDirectory.load_config_paths('bmpc'):
            try:
                self.config = self.config or json.load(open(os.path.join(path, 'config.json')))
            except Exception as e:
                logger.warning(e)

        self.config.setdefault('server', {})
        self.config['server'].setdefault('profiles',
                                         [
                                             {
                                                 'name': _('Local host'),
                                                 'host': 'localhost',
                                                 'port': 6600,
                                             },
                                         ])
        self.config['server'].setdefault('profile', _('Local host'))

    def close(self):
        logger.debug('Closing shared')
        s = json.dumps(self.config, sort_keys=True, indent=2, ensure_ascii=False)
        open(os.path.join(xdg.BaseDirectory.save_config_path('bmpc'), 'config.json'), 'w').write(s)
        self.ampd_client.close()
        sys.excepthook = self.excepthook_orig
        del self.plugins

    def build_ui(self, uiname):
        if uiname in self.uidefs:
            uidef = self.uidefs[uiname]
        else:
            uidef = open(os.path.join(os.path.dirname(__file__), uiname + '.ui')).read()
            self.uidefs[uiname] = uidef
        builder = Gtk.Builder(translation_domain='bmpc')
        builder.add_from_string(uidef)
        return builder

    def connect_cb(self):
        logger.info(_('Connected [protocol version {}]').format(self.ampd_worker_group.get_protocol_version()))
        self.idle_player()
        self.idle_options()
        self.idle_database()
        self.idle_mixer()
        self.playtime_update()

    def disconnect_cb(self):
        logger.info(_('Disconnected'))
        self.current_song = {}
        self.state = ''
        self.playtime_upper = 0
        self.playtime_value = 0
        self.reconnect()

    def connect_failed_cb(self):
        logger.info(_('Connection failed'))
        self.reconnect()

    @ampd.worker
    def reconnect(self):
        if self.want_to_connect:
            reason = yield ampd.CONNECT, ampd.TIMEOUT(1000)
            if ampd.CONNECT not in reason:
                self.ampd_client.connect(self.ampd_host, self.ampd_port)

    @ampd.worker
    def idle_player(self):
        while True:
            status = yield ampd.status()
            self.state = status['state']
            if self.state == 'stop':
                self.current_song = {}
            else:
                song = yield ampd.currentsong()
                self.current_song = song
            if 'error' in status:
                logger.error(status['error'])
                yield ampd.clearerror()
            yield ampd.PLAYER

    @ampd.worker
    def idle_mixer(self):
        while True:
            status = yield ampd.status()
            self.handler_block_by_func(self.notify_volume_cb)
            self.volume = int(status['volume'])
            self.handler_unblock_by_func(self.notify_volume_cb)
            yield ampd.MIXER

    @staticmethod
    @ampd.worker
    def notify_volume_cb(self, param):
        while True:
            try:
                yield ampd.setvol(self.volume)
                return
            except:
                yield ampd.PLAYER

    @ampd.worker
    def playtime_update(self):
        while True:
            status = yield ampd.status()
            self.bitrate = status.get('bitrate')
            if 'time' in status:
                times = status['time'].split(':')
                self.playtime_upper = int(times[1])
                self.handler_block_by_func(self.notify_playtime_value_cb)
                self.playtime_value = int(times[0])
                self.handler_unblock_by_func(self.notify_playtime_value_cb)
            else:
                self.playtime_upper = 0
                self.playtime_value = 0
            if status['state'] == 'play':
                yield ampd.PLAYER, ampd.TIMEOUT((self.playtime_value + 1 - float(status['elapsed'])) * 1000)
            else:
                yield ampd.PLAYER

    @staticmethod
    @ampd.worker
    def notify_playtime_value_cb(self, param):
        if self.playtime_upper:
            yield ampd.seekcur(self.playtime_value)

    @ampd.worker
    def idle_options(self):
        while True:
            status = yield ampd.status()
            for option in self.option_names:
                value = bool(int(status[option]))
                if value != self.get_property('option-' + option):
                    if self.safe_mode:
                        yield ampd.commands_[option](0)
                    else:
                        self.set_property('option-' + option, value)
            yield ampd.OPTIONS

    @staticmethod
    @ampd.worker
    def notify_option_cb(self, param):
        option = param.name.split('-')[1]
        yield ampd.commands_[option](int(self.get_property(param.name)))

    @ampd.worker
    def idle_database(self):
        while True:
            idle = yield ampd.DATABASE, ampd.UPDATE
            status = yield ampd.status()
            if 'updating_db' in status:
                logger.info(_('Updating database'))
            elif ampd.DATABASE in idle:
                logger.info(_('Database changed'))
            else:
                logger.info(_('Database unchanged'))

    @ampd.worker
    def mpd_command_cb(self, caller, *data):
        if not self.state:
            yield ampd.CONNECT
        yield ampd.commands_[caller.get_name()]()

    @ampd.worker
    def play_or_pause_cb(self, caller, *data):
        if not self.state:
            yield ampd.CONNECT
            status = yield ampd.status()
            self.state = status['state']
        yield ampd.pause(1) if self.state == 'play' else ampd.play()

    @ampd.worker
    def fade_to_next_cb(self, caller, *data):
        N = 30
        T = 7

        if not self.state:
            yield ampd.IDLE
        if self.state != 'play':
            return
        volume = self.volume
        for i in range(N):
            self.volume = volume*(N-i-1)/N
            yield ampd.TIMEOUT(T*1000/N)
        yield ampd.next()
        self.volume = volume

    @staticmethod
    def format_time(time):
        time = int(time)
        hours = time // 3600
        minutes = (time // 60) % 60
        seconds = time % 60
        return '{:d}:{:02d}:{:02d}'.format(hours, minutes, seconds) if hours else '{:d}:{:02d}'.format(minutes, seconds)

    def ampd_connect(self, action, state):
        self.want_to_connect = True
        if state:
            action.set_state(state)
            self.config['server']['profile'] = state.get_string()
        name = self.config['server']['profile']
        for profile in self.config['server']['profiles']:
            if profile['name'] == name:
                self.ampd_host = profile['host']
                self.ampd_port = profile['port']
                self.ampd_client.connect(self.ampd_host, self.ampd_port)
                break

    def ampd_disconnect(self, *args):
        self.want_to_connect = False
        self.ampd_client.disconnect()

    def excepthook(self, *args):
        logger.error('EXCEPTION', exc_info=args)

    @staticmethod
    def sqlite_connection(name):
        connection = apsw.Connection(os.path.join(xdg.BaseDirectory.save_data_path('bmpc'), name + '.sqlite'))
        connection.cursor().execute('PRAGMA FOREIGH_KEYS=ON')
        return connection
