# coding: utf-8
#
# BMPC Music Player Client
#
# Copyright (C) 2015 Itaï BEN YAACOV
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from gi.repository import Gio, GLib


class OrderedMenu(Gio.Menu):
    ATTRIBUTE_NAME = 'omenu-name'

    @classmethod
    def _create_item(cls, name, target, label=None, attributes=None):
        item = Gio.MenuItem.new(label)
        item.set_attribute_value(cls.ATTRIBUTE_NAME, GLib.Variant.new_string(name))
        if target in (Gio.MENU_LINK_SUBMENU, Gio.MENU_LINK_SECTION):
            submenu = cls()
            item.set_link(target, submenu)
        else:
            submenu = None
            item.set_detailed_action(target)
            if attributes:
                for attribute, value in attributes.items():
                    if value == None:
                        continue
                    elif isinstance(value, str):
                        value = GLib.Variant.new_string(value)
                    else:
                        value = GLib.Variant.new_strv(value)
                    item.set_attribute_value(attribute.replace('_', '-'), value)
        return item, submenu

    def _find_position(self, name):
        n = self.get_n_items()
        for i in range(n):
            item_name = self.get_item_attribute_value(i, self.ATTRIBUTE_NAME)
            item_name = item_name and item_name.get_string()
            if item_name == None or item_name >= name:
                return i, item_name == name
        return n, False

    def insert_path(self, path, target, label=None, accels=None, **attributes):
        '''
        path = NAME/NAME/..../NAME
        target = Gio.MENU_LINK_SUBMENU | Gio.MENU_LINK_SECTION | detailed_action
        '''

        menu = self
        names = path.split('/')
        for name in names[:-1]:
            i, found = menu._find_position(name)
            if not found:
                item, submenu = menu._create_item(name, Gio.MENU_LINK_SECTION)
                menu.insert_item(i, item)
                menu = submenu
            else:
                menu = menu.get_item_link(i, Gio.MENU_LINK_SUBMENU) or menu.get_item_link(i, Gio.MENU_LINK_SECTION)
                if not menu:
                    raise ValueError('Path element "{}" is not a submenu or section'.format(name))

        name = names[-1]
        i, found = menu._find_position(name)
        if found:
            if target in (Gio.MENU_LINK_SUBMENU, Gio.MENU_LINK_SECTION):
                submenu = menu.get_item_link(i, target)
                if submenu:
                    return submenu
            raise ValueError('Item already exists for path "{}"'.format(path))
        attributes.setdefault('hidden_when', 'action-missing')
        if accels:
            attributes['accels'] = accels
        item, submenu = menu._create_item(name, target, label, attributes)
        menu.insert_item(i, item)
        return submenu

    def get_accels(self):
        accels = {}
        for i in range(self.get_n_items()):
            submenu = self.get_item_link(i, Gio.MENU_LINK_SUBMENU) or self.get_item_link(i, Gio.MENU_LINK_SECTION)
            if submenu:
                accels.update(submenu.get_accels())
            else:
                item_accels = self.get_item_attribute_value(i, 'accels')
                if not item_accels:
                    continue
                item_action = self.get_item_attribute_value(i, 'action').get_string()
                item_target = self.get_item_attribute_value(i, 'target')
                accels[Gio.action_print_detailed_name(item_action, item_target)] = item_accels.get_strv()
        return accels
