# coding: utf-8
#
# BMPC Music Player Client
#
# Copyright (C) 2015 Itaï BEN YAACOV
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from gi.repository import GObject, GLib, Gtk, Gdk
import logging

from .utils import actions
from .utils.logger import logger


class WindowLoggingHandeler(logging.Handler):
    def __init__(self, widget, timeout):
        super(WindowLoggingHandeler, self).__init__(logging.INFO)
        self.widget = widget
        self.timeout = None
        self.timeout_time = timeout

    def emit(self, record):
        label = self.format(record)
        #self.label_label.set_markup('<span background="red" size="x-large">{}</span>'.format(label))
        self.widget.set_label(label)
        self.widget.set_visible(True)
        if self.timeout:
            GLib.source_remove(self.timeout)
        self.timeout = GLib.timeout_add(self.timeout_time, self.reset_label)

    def reset_label(self):
        self.timeout = None
        self.widget.set_visible(False)

    def close(self):
        if self.timeout:
            GLib.source_remove(self.timeout)
        super(WindowLoggingHandeler, self).close()


class Window(Gtk.ApplicationWindow, actions.ActionGroup):
    def __init__(self, app):
        super(Window, self).__init__(application=app)
        self.set_wmclass('bmpc', 'BMPC')
        self.set_title('BMPC')
        self.shared = app.shared

        self.config = self.shared.config.setdefault('window', {})
        self.config.setdefault('message_timeout', 5)
        self.set_default_size(self.config.setdefault('width', 1000),
                              self.config.setdefault('height', 600))

        self.show()

        self.add_complex_action('toggle-fullscreen', self.action_toggle_fullscreen_cb)

        self.connect('destroy', self.destroy_cb)
        self.connect('check-resize', self.check_resize_cb)
        self.connect('window-state-event', self.window_state_event_cb)

        self.shared.connect('notify::current-song', self.notify_current_song_cb)

        self.mpd_command_cb = self.shared.mpd_command_cb
        builder = self.shared.build_ui('window')
        builder.connect_signals(self)

        self.option_buttons = {}
        for option in self.shared.option_names:
            button = builder.get_object('togglebutton-' + option)
            self.shared.bind_property('option-' + option, button, 'active', GObject.BindingFlags.BIDIRECTIONAL | GObject.BindingFlags.SYNC_CREATE)
            self.shared.bind_property('safe-mode', button, 'sensitive', GObject.BindingFlags.SYNC_CREATE, lambda x, y: not y)
            self.option_buttons[option] = button
            self.option_notify_active_cb(button, None)
            button.connect('notify::active', self.option_notify_active_cb)

        self.scale_time = builder.get_object('scale-time')
        label_time = builder.get_object('label-time')
        label_time_total = builder.get_object('label-time-total')
        self.volume_button = builder.get_object('volume-button')
        self.label_component = builder.get_object('label-component')
        self.box_status = builder.get_object('box-status')
        self.box = builder.get_object('box')
        self.add(self.box)

        self.shared.bind_property('state', builder.get_object('button-play'), 'visible', GObject.BindingFlags.SYNC_CREATE, lambda x, y: y != 'play')
        self.shared.bind_property('state', builder.get_object('button-pause'), 'visible', GObject.BindingFlags.SYNC_CREATE, lambda x, y: y == 'play')
        self.shared.bind_property('state', self.volume_button, 'sensitive', GObject.BindingFlags.SYNC_CREATE, lambda x, y: y != 'stop')
        self.shared.bind_property('volume', self.volume_button, 'value', GObject.BindingFlags.BIDIRECTIONAL | GObject.BindingFlags.SYNC_CREATE)
        self.shared.bind_property('safe-mode', builder.get_object('image-safe-mode'), 'visible', GObject.BindingFlags.SYNC_CREATE)
        self.shared.bind_property('safe-mode', self.scale_time, 'sensitive', GObject.BindingFlags.SYNC_CREATE, lambda x, y: not y)
        label_time.bind_property('visible', label_time_total, 'visible', GObject.BindingFlags.SYNC_CREATE)
        self.shared.bind_property('bitrate', builder.get_object('label-bitrate'), 'label', GObject.BindingFlags.SYNC_CREATE, lambda x, y: ' [{}kbps]'.format(y) if y else '')
        self.shared.bind_property('playtime-upper', self.scale_time.get_adjustment(), 'upper', GObject.BindingFlags.SYNC_CREATE)
        self.shared.bind_property('playtime-value', self.scale_time, 'fill-level', GObject.BindingFlags.SYNC_CREATE)
        self.scale_time.get_adjustment().bind_property('value', label_time, 'label', GObject.BindingFlags.SYNC_CREATE, lambda x, y: self.shared.format_time(y))
        self.scale_time.get_adjustment().bind_property('upper', label_time_total, 'label', GObject.BindingFlags.SYNC_CREATE, lambda x, y: ' / ' + self.shared.format_time(y))
        self.scale_time.get_adjustment().bind_property('upper', label_time, 'visible', GObject.BindingFlags.SYNC_CREATE, lambda x, y: y != 0.0)
        self.scale_time_binding = None
        self.set_scale_time_binding()

        self.is_fullscreen = False

        self.component = None

        self.logging_handler = WindowLoggingHandeler(builder.get_object('label-message'), self.config['message_timeout'] * 1000)
        logger.addHandler(self.logging_handler)

    @staticmethod
    def destroy_cb(self):
        logger.debug('Destroying window: {}'.format(self))
        self.shared.disconnect_by_func(self.notify_current_song_cb)
        self.change_component(None)
        logger.removeHandler(self.logging_handler)
        self.logging_handler.close()

    def option_notify_active_cb(self, button, param):
        option_name = '{} mode'.format(button.get_name().capitalize())
        button.set_tooltip_text('{} {}'.format(_(option_name), _('on') if button.get_active() else _('off')))

    def change_component(self, component):
        if self.component:
            self.component.stop()
            self.box.remove(self.component.widget)
        self.component = component
        if component:
            self.box.pack_start(component.widget, True, True, 0)
            component.start(self)
            self.update_title()
        else:
            self.set_title('BMPC')

    def notify_current_song_cb(self, *args):
        self.scale_time_break_interaction()
        self.update_title()

    def update_title(self):
        title = 'BMPC'
        if self.component:
            title += ' - {}'.format(self.component.title)
        if self.shared.state != 'stop':
            title += ' - {}'.format(self.shared.current_song.get('title', ''))
            if self.shared.state == 'pause':
                title += ' ({})'.format('paused')
        self.set_title(title)

    @staticmethod
    def check_resize_cb(self):
        if not self.is_fullscreen:
            self.config['width'], self.config['height'] = self.get_size()

    def scale_time_button_press_event_cb(self, scale_time, event):
        if event.button == 1 and event.type == Gdk.EventType.BUTTON_PRESS and self.scale_time_binding:
            self.scale_time_binding.unbind()
            self.scale_time_binding = None
        else:
            self.scale_time_break_interaction()

    def scale_time_button_release_event_cb(self, scale_time, event):
        if not self.scale_time_binding:
            if event.button == 1 and event.state & (Gdk.ModifierType.SHIFT_MASK | Gdk.ModifierType.CONTROL_MASK | Gdk.ModifierType.MOD1_MASK) == 0:
                self.shared.playtime_value = self.scale_time.get_value()
                self.set_scale_time_binding()
            else:
                self.scale_time_break_interaction()

    def scale_time_break_interaction(self, *args):
        if not self.scale_time_binding:
            self.scale_time.set_sensitive(False)
            GLib.timeout_add(0, self.scale_time.set_sensitive, True)

    def scale_time_notify_sensitive_cb(self, *args):
        if not self.scale_time.get_sensitive():
            self.set_scale_time_binding()

    def set_scale_time_binding(self):
        if not self.scale_time_binding:
            self.scale_time_binding = self.shared.bind_property('playtime-value', self.scale_time.get_adjustment(), 'value', GObject.BindingFlags.SYNC_CREATE)

    def action_toggle_fullscreen_cb(self, *args):
        if self.is_fullscreen:
            self.unfullscreen()
        else:
            self.fullscreen()

    @staticmethod
    def window_state_event_cb(self, event):
        self.is_fullscreen = bool(event.new_window_state & Gdk.WindowState.FULLSCREEN)
        self.set_show_menubar(not self.is_fullscreen)
        self.box_status.set_visible(not self.is_fullscreen)
        return False
